import pytest

import numpy as np
import capytaine as cpt

from capytaine.tools.prony_decomposition import exponential_decomposition, find_best_exponential_decomposition

faces_examples = {
        "immersed": cpt.Mesh(vertices=[[0.0, 0.0, -1.0], [1.0, 0.0, -1.0], [1.0, 1.0, -1.0], [0.0, 1.0, -1.0]], faces=np.array([[0, 1, 2, 3]])),
        "free_surface": cpt.Mesh(vertices=[[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [0.0, 1.0, 0.0]], faces=np.array([[0, 1, 2, 3]])),
        }

@pytest.mark.parametrize("derivative_with_respect_to_first_variable", [True, False])
@pytest.mark.parametrize("face_location", ["immersed", "free_surface"])
def test_deep_water_asymptotics_of_wave_terms_low_freq(derivative_with_respect_to_first_variable, face_location):
    gf = cpt.Delhommeau()
    gf_singularities_index = gf.gf_singularities_fortran_enum["low_freq"]
    face = faces_examples[face_location]
    k = 1.0
    depth = 1000.0
    s_inf, k_inf = gf.fortran_core.green_wave.integral_of_wave_part_infinite_depth(
        face.faces_centers[0, :], face.faces_centers[0, :], face.faces_areas[0], face.quadrature_points[0][0, :, :], face.quadrature_points[1][0],
        k, *gf.all_tabulation_parameters, gf_singularities_index, derivative_with_respect_to_first_variable
    )
    s_finite, k_finite = gf.fortran_core.green_wave.integral_of_wave_parts_finite_depth(
        face.faces_centers[0, :], face.faces_centers[0, :], face.faces_areas[0], face.quadrature_points[0][0, :, :], face.quadrature_points[1][0],
        k, depth, *gf.all_tabulation_parameters, gf_singularities_index, derivative_with_respect_to_first_variable
    )
    np.testing.assert_allclose(s_inf, s_finite, rtol=1e-2)
    np.testing.assert_allclose(k_inf, k_finite, rtol=1e-2)


@pytest.mark.parametrize("derivative_with_respect_to_first_variable", [True, False])
@pytest.mark.parametrize("face_location", ["immersed"])
def test_deep_water_asymptotics_of_wave_terms_high_freq(derivative_with_respect_to_first_variable, face_location):
    gf = cpt.Delhommeau()
    gf_singularities_index = gf.gf_singularities_fortran_enum["high_freq"]
    face = faces_examples[face_location]
    k = 1.0
    depth = 1000.0
    s_inf, k_inf = gf.fortran_core.green_wave.integral_of_wave_part_infinite_depth(
        face.faces_centers[0, :], face.faces_centers[0, :], face.faces_areas[0], face.quadrature_points[0][0, :, :], face.quadrature_points[1][0],
        k, *gf.all_tabulation_parameters, gf_singularities_index, derivative_with_respect_to_first_variable
    )
    s_finite, k_finite = gf.fortran_core.green_wave.integral_of_wave_parts_finite_depth(
        face.faces_centers[0, :], face.faces_centers[0, :], face.faces_areas[0], face.quadrature_points[0][0, :, :], face.quadrature_points[1][0],
        k, depth, *gf.all_tabulation_parameters, gf_singularities_index, derivative_with_respect_to_first_variable
    )
    np.testing.assert_allclose(s_inf, s_finite, rtol=1e-2)
    np.testing.assert_allclose(k_inf, k_finite, rtol=1e-2)


@pytest.mark.parametrize("derivative_with_respect_to_first_variable", [True, False])
@pytest.mark.parametrize("face_location", ["immersed", "free_surface"])
def test_deep_water_asymptotics_of_prony_decomposition(derivative_with_respect_to_first_variable, face_location):
    gf = cpt.Delhommeau()
    k = 1.0
    depth = 1000.0
    face = faces_examples[face_location]
    decomp = gf.find_best_exponential_decomposition(k*depth, method="python")
    s_prony, k_prony = gf.fortran_core.green_wave.integral_of_prony_decomp_finite_depth(
            face.faces_centers[0, :], face.vertices[face.faces[0, :], :], face.faces_centers[0, :], face.faces_normals[0, :], face.faces_areas[0], face.faces_radiuses[0],
            depth, decomp, derivative_with_respect_to_first_variable)
    np.testing.assert_allclose(s_prony, 0.0, atol=1e-3)
    np.testing.assert_allclose(k_prony, 0.0, atol=1e-3)


def test_high_freq_vs_low_freq():
    gf = cpt.Delhommeau()
    mesh = cpt.mesh_sphere().immersed_part()
    k = 1.0
    depth = 10.0
    decomp = gf.find_best_exponential_decomposition(k*depth, method="python")
    def compute_matrices(gf_singularities):
        adjoint_double_layer = True
        S, K = gf._init_matrices(
            (mesh.nb_faces, mesh.nb_faces), early_dot_product=True
        )
        gf.fortran_core.matrices.build_matrices(
                mesh.faces_centers, mesh.faces_normals,
                mesh.vertices, mesh.faces + 1,
                mesh.faces_centers, mesh.faces_normals,
                mesh.faces_areas, mesh.faces_radiuses,
                *mesh.quadrature_points,
                k, depth,
                *gf.all_tabulation_parameters,
                gf.finite_depth_method_index, decomp,
                gf.dispersion_relation_roots,
                gf.gf_singularities_fortran_enum[gf_singularities],
                adjoint_double_layer,
                S, K,
                )
        return S, K
    S_low, K_low = compute_matrices('low_freq')
    S_high, K_high = compute_matrices('high_freq')
    np.testing.assert_allclose(S_low, S_high, atol=1e-2*np.linalg.norm(S_low))
    np.testing.assert_allclose(K_low, K_high, atol=2e-2*np.linalg.norm(K_low))



def test_infinite_frequency_prony_decomposition():
    gf = cpt.Delhommeau()
    with pytest.raises(NotImplementedError):
        gf.find_best_exponential_decomposition(np.inf, method='fortran')
    gf.find_best_exponential_decomposition(np.inf, method='python')


def test_infinite_frequency():
    gf = cpt.Delhommeau(gf_singularities='high_freq')
    mesh = cpt.mesh_sphere().immersed_part()
    S_, K_ = gf.evaluate(mesh, mesh, free_surface=0.0, water_depth=10.0, wavenumber=1000.0)
    S, K = gf.evaluate(mesh, mesh, free_surface=0.0, water_depth=10.0, wavenumber=np.inf)
    np.testing.assert_allclose(S/np.linalg.norm(S), S_/np.linalg.norm(S_), atol=1e-3)
    np.testing.assert_allclose(K/np.linalg.norm(K), K_/np.linalg.norm(K_), atol=1e-3)


def test_prony_decomposition():
    x = np.linspace(0.0, 1.0, 100)
    y = 2*np.exp(-x) - 4*np.exp(-2*x)
    a, lamda = exponential_decomposition(x, y, 2)
    assert ((np.allclose(a, [-4.0, 2.0]) and np.allclose(lamda, [-2.0, -1.0]))
            or (np.allclose(a, [2.0, -4.0]) and np.allclose(lamda, [-1.0, -2.0])))


def test_find_best_prony_decomposition():
    rng = np.random.default_rng(seed=987654321)
    # May fail for other RNGs, for instance if two values from lamda_refs are very close from each other
    n_ref = 6
    a_ref = rng.uniform(-10.0, 10.0, size=n_ref)
    lamda_ref = -rng.chisquare(1.0, size=n_ref)
    def f(x):
        return sum(a_ref[i]*np.exp(lamda_ref[i]*x) for i in range(n_ref))
    a, lamda = find_best_exponential_decomposition(f, 0.0, 100.0, range(2, 15), tol=1e-4, noise_on_domain_points_std=0.0)
    assert np.allclose(np.sort(a), np.sort(a_ref))
    assert np.allclose(np.sort(lamda_ref), np.sort(lamda_ref))


def test_failure_for_low_kh():
    gf = cpt.Delhommeau()

    # Legacy method returning crappy value
    gf.find_best_exponential_decomposition(0.10, method="fortran")

    # Newer method fails properly
    from capytaine.green_functions.abstract_green_function import GreenFunctionEvaluationError
    with pytest.raises((NotImplementedError, GreenFunctionEvaluationError)):
        gf.find_best_exponential_decomposition(0.10, method="python")
    with pytest.raises((NotImplementedError, GreenFunctionEvaluationError)):
        gf.find_best_exponential_decomposition(0.01, method="python")


@pytest.mark.xfail  # Find a way to pass parameters to find_best_exponential_decomposition to make sure they match
def test_python_and_fortran_prony_decomposition_for_green_function():
    gf = cpt.Delhommeau(finite_depth_prony_decomposition_method="python")
    decomp_default = gf.find_best_exponential_decomposition(1.0)
    decomp_f = gf.find_best_exponential_decomposition(1.0, method="fortran")
    decomp_p = gf.find_best_exponential_decomposition(1.0, method="python")
    assert np.allclose(decomp_default, decomp_p)
    assert np.allclose(decomp_p, decomp_f, rtol=0.2)


def test_prony_decomposition_specific_value():
    gf = cpt.Delhommeau(finite_depth_prony_decomposition_method="python")
    gf.find_best_exponential_decomposition(20.0)


def test_failure_unknown_method():
    gf = cpt.Delhommeau()
    with pytest.raises(ValueError):
        gf.find_best_exponential_decomposition(1.0, method="potato")


def test_fingreen3D():
    k = 5.0
    h = 2.0
    mesh = cpt.mesh_sphere(radius=0.5, center=(0.0, 0.0, -1.0), resolution=(4, 4))
    S, D = cpt.FinGreen3D().evaluate(mesh, mesh, free_surface=0.0, water_depth=h, wavenumber=k, adjoint_double_layer=False)
    S_ref, D_ref = cpt.Delhommeau().evaluate(mesh, mesh, free_surface=0.0, water_depth=h, wavenumber=k, adjoint_double_layer=False)
    np.testing.assert_allclose(S, S_ref, rtol=1e-1, atol=1e-1)
    np.testing.assert_allclose(D, D_ref, rtol=1e-1, atol=1e-1)


@pytest.mark.parametrize("omega", [0.1, 1.0, 10.0])
@pytest.mark.parametrize("depth", [0.1, 1.0, 10.0])
def test_fingreen3D_roots(omega, depth):
    from scipy.optimize import newton, brentq
    nk = 100
    omega2_over_g = omega**2/9.81
    all_roots = cpt.FinGreen3D().fortran_core.green_wave.compute_dispersion_roots(nk, omega2_over_g, depth)
    wavenumber = newton(lambda x: omega2_over_g - x*np.tanh(x*depth), omega2_over_g)
    assert np.isclose(wavenumber, all_roots[0], rtol=1e-3)

    roots_idempotent = newton(lambda x: -omega2_over_g - x*np.tan(x*depth), all_roots[1:])
    np.testing.assert_allclose(roots_idempotent, all_roots[1:], rtol=1e-3)

    def root(i_root):
        return brentq(lambda y: omega2_over_g*depth + y*np.tan(y), (2*i_root+1)*np.pi/2 + 1e-10, (2*i_root+2)*np.pi/2 - 1e-10)/depth
    roots_2 = [root(i_root) for i_root in range(nk-1)]
    np.testing.assert_allclose(roots_2, all_roots[1:], rtol=1e-4)

    all_roots_python = cpt.FinGreen3D().compute_dispersion_relation_roots(nk, wavenumber, depth)
    np.testing.assert_allclose(all_roots_python, all_roots, rtol=1e-3)



import pytest

import numpy as np
import capytaine as cpt

from capytaine.bodies.predefined.rectangles import Rectangle, OpenRectangularParallelepiped, RectangularParallelepiped
from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import Disk, HorizontalCylinder, VerticalCylinder


def test_rectangle_generation():
    # Rejects invalid inputs:
    with pytest.raises(AssertionError):
        Rectangle(size=(5.0, 6.0, 7.0))

    with pytest.raises(AssertionError):
        Rectangle(size=(-4, 5))

    with pytest.raises(AssertionError):
        Rectangle(resolution=(4.1, 5))

    with pytest.raises(ValueError):
        Rectangle(resolution=(3, 3), reflection_symmetry=True)

    # Test mesh with symmetry
    rec = Rectangle(size=(10, 10), resolution=(6, 2), normal=(1, 0, 0),
                    reflection_symmetry=False, translational_symmetry=False,
                    center=(0, 0, -5), name="test")

    assert rec.name == "test"
    assert isinstance(rec.mesh, cpt.Mesh)
    assert rec.mesh.name == "test_mesh"
    assert all(rec.size == [10.0, 10.0])
    assert all(rec.geometric_center == [0, 0, -5.0])
    assert rec.mesh.nb_faces == 12
    assert rec.mesh.nb_vertices == 21

    # x coordinate
    assert np.allclose(rec.mesh.vertices[:, 0], 0.0)

    # y coordinate
    assert np.all(rec.mesh.vertices[:, 1] <= 5.0)
    assert np.all(rec.mesh.vertices[:, 1] >= -5.0)

    # z coordinate
    assert np.all(rec.mesh.vertices[:, 2] <= 0.0)
    assert np.all(rec.mesh.vertices[:, 2] >= -10.0)

    # Test mesh with reflection symmetry
    sym_rec = Rectangle(size=(10, 10), resolution=(6, 2),
                        translational_symmetry=False, reflection_symmetry=True,
                        center=(0, 0, -5))
    assert isinstance(sym_rec.mesh, cpt.ReflectionSymmetricMesh)
    assert sym_rec.mesh.nb_submeshes == 2
    assert sym_rec.mesh.nb_faces == 12

    sym_rec_merged_mesh = sym_rec.mesh.merged()
    assert isinstance(sym_rec_merged_mesh, cpt.Mesh)
    # assert sym_rec_merged_mesh == rec.mesh

    # Test mesh with translation symmetry
    trans_rec = Rectangle(size=(10, 10), resolution=(6, 2),
                          translational_symmetry=True, reflection_symmetry=False,
                          center=(0, 0, -5))
    assert isinstance(trans_rec.mesh, cpt.TranslationalSymmetricMesh)
    assert trans_rec.mesh.nb_submeshes == 6
    assert trans_rec.mesh.nb_faces == 12

    trans_rec_merged_mesh = trans_rec.mesh.merged()
    assert isinstance(trans_rec_merged_mesh, cpt.Mesh)
    # assert trans_rec_merged_mesh == rec.mesh


def test_open_parallelepiped_generation():
    para = OpenRectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                         resolution=(5, 2, 3),
                                         translational_symmetry=False, name="test")
    assert para.name == "test"
    assert isinstance(para.mesh, cpt.Mesh)
    assert para.mesh.name == "test_mesh"
    assert para.mesh.nb_faces == 42

    assert np.isclose(np.abs(para.mesh.vertices[:, 0]).max(), 2.5)
    assert np.isclose(np.abs(para.mesh.vertices[:, 1]).max(), 0.5)
    assert np.all(para.mesh.vertices[:, 2] <= 0.0)
    assert np.all(para.mesh.vertices[:, 2] >= -3.0)

    clever_para = OpenRectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                                resolution=(5, 2, 3),
                                                translational_symmetry=True, name="clever_test")

    assert clever_para.mesh.nb_faces == 42
    assert isinstance(clever_para.mesh, cpt.CollectionOfMeshes)
    assert any(isinstance(submesh, cpt.TranslationalSymmetricMesh) for submesh in clever_para.mesh)


def test_parallelepiped_generation():
    full_para = RectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                          resolution=(6, 2, 3),
                                          translational_symmetry=False, name="full_test")
    assert full_para.mesh.nb_faces == 72
    # full_para.show()

    sym_para = RectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                         resolution=(6, 2, 3),
                                         reflection_symmetry=True, name="sym_full_test")
    assert sym_para.mesh.nb_faces == 72
    # sym_para.show()

    trans_para = RectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                           resolution=(6, 2, 3),
                                           translational_symmetry=True, name="trans_full_test")
    assert trans_para.mesh.nb_faces == 72
    # trans_para.show()


@pytest.mark.parametrize(
        "shape", [RectangularParallelepiped, HorizontalCylinder, VerticalCylinder, Sphere]
        )
def test_geometric_center(shape):
    center = (10, 5, -1.5)
    body = shape(center=center)
    assert hasattr(body, "geometric_center")
    body.add_all_rigid_body_dofs()
    assert np.allclose(body._infer_rotation_center(), center)


def test_disk():
    a = Disk(resolution=(6, 6))
    b = Disk(resolution=(6, 6), axial_symmetry=True)
    # print(b.mesh.tree_view())
    c = Disk(resolution=(6, 6), reflection_symmetry=True)
    # TODO

def test_cylinder():
    HorizontalCylinder()
    VerticalCylinder()
    # TODO


def test_cylinder_submesh_names():
    for s in [True, False]:
        c = HorizontalCylinder(reflection_symmetry=True, translation_symmetry=s, name="cylinder")
        assert c.mesh[0].name == "half_cylinder_mesh"
        assert c.mesh[1].name == "mirrored_of_half_cylinder_mesh"


############
#  SPHERE  #
############

def test_sphere_name():
    sphere = Sphere()
    assert sphere.name.startswith("sphere_")

def test_sphere_axisymmetric():
    sphere = Sphere(axial_symmetry=True)
    assert isinstance(sphere.mesh, cpt.AxialSymmetricMesh)

def test_sphere_not_axisymmetric():
    sphere = Sphere(axial_symmetry=False)
    assert isinstance(sphere.mesh, cpt.Mesh)

def test_sphere_nb_panels():
    sphere = Sphere(ntheta=5, nphi=5, clip_free_surface=False)
    assert sphere.mesh.nb_faces == 25

def test_sphere_nb_panels_clipped():
    sphere = Sphere(ntheta=5, nphi=5, clip_free_surface=True)
    assert sphere.mesh.nb_faces == 25

def test_sphere_nb_panels_clipped_is_underwater():
    sphere = Sphere(ntheta=5, nphi=5, clip_free_surface=True)
    assert np.all(sphere.mesh.vertices[:, 2] <= 0.0)

def test_sphere_out_of_water():
    with pytest.raises(ValueError):
        sphere = Sphere(radius=1.0, center=(0, 0, 10), clip_free_surface=True)

def test_sphere_geometric_center():
    sphere = Sphere(radius=2.0, center=(-2.0, 2.0, 1.0))
    assert np.allclose(sphere.geometric_center, np.array([-2.0, 2.0, 1.0]))

def test_sphere_clipping():
    s1 = Sphere(ntheta=4, nphi=4, axial_symmetry=False, clip_free_surface=True)
    s2 = Sphere(ntheta=8, nphi=4, axial_symmetry=False, clip_free_surface=False).keep_immersed_part()
    assert s1.mesh.nb_faces == s2.mesh.nb_faces
    assert s1.mesh.nb_vertices == s2.mesh.nb_vertices
    # TODO: test that the faces are actually the same.



import pytest

import numpy as np
import xarray as xr

import capytaine as cpt

from capytaine.io.xarray import problems_from_dataset, separate_complex_values, merge_complex_values


@pytest.fixture
def sphere():
    sphere = cpt.FloatingBody(
            mesh=cpt.mesh_sphere(center=(0, 0, -2), radius=1.0, resolution=(10, 20)),
            name="sphere",
            )
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    return sphere


@pytest.fixture
def solver():
    solver = cpt.BEMSolver()
    return solver


#######################################################################
#                       Problems from datasets                        #
#######################################################################

def test_problems_from_dataset(sphere):
    dset = xr.Dataset(coords={'omega': [0.5, 1.0, 1.5],
                              'radiating_dof': ["Heave"],
                              'body_name': ["sphere"],
                              'wave_direction': [0.0],
                              'water_depth': [np.inf]})

    problems = problems_from_dataset(dset, [sphere])
    assert cpt.RadiationProblem(body=sphere, omega=0.5, radiating_dof="Heave") in problems
    assert len(problems) == 6
    assert len([problem for problem in problems if isinstance(problem, cpt.DiffractionProblem)]) == 3

    dset = xr.Dataset(coords={'omega': [0.5, 1.0, 1.5],
                              'wave_direction': [0.0],
                              'body_name': ["cube"]})
    with pytest.raises(AssertionError):
        problems_from_dataset(dset, [sphere])

    shifted_sphere = sphere.translated_y(5.0, name="shifted_sphere")
    dset = xr.Dataset(coords={'omega': [0.5, 1.0, 1.5],
                              'radiating_dof': ["Heave"],
                              'wave_direction': [0.0]})
    problems = problems_from_dataset(dset, [sphere, shifted_sphere])
    assert cpt.RadiationProblem(body=sphere, omega=0.5, radiating_dof="Heave") in problems
    assert cpt.RadiationProblem(body=shifted_sphere, omega=0.5, radiating_dof="Heave") in problems
    assert len(problems) == 12


def test_problems_from_dataset_incomplete_test_matrix():
    body = cpt.FloatingBody(cpt.mesh_sphere())
    test_matrix = xr.Dataset(coords={
        "omega": np.linspace(0, 1, 2),
        })
    with pytest.raises(ValueError):
        problems_from_dataset(test_matrix, body)


def test_problems_from_dataset_with_wavelength(sphere):
    dset = xr.Dataset(coords={'wavelength': [12.0], 'radiating_dof': ["Heave"]})
    problems = problems_from_dataset(dset, sphere)
    for pb in problems:
        assert np.isclose(pb.wavelength, 12.0)


def test_problems_from_dataset_with_too_many_frequencies(sphere):
    dset = xr.Dataset(coords={'wavelength': [12.0], 'period': [3.0], 'radiating_dof': ["Heave"]})
    with pytest.raises(ValueError, match="at most one"):
        problems_from_dataset(dset, sphere)


def test_problems_from_dataset_without_list(sphere):
    dset = xr.Dataset(coords={'omega': 1.5, 'radiating_dof': "Heave"})
    problems = problems_from_dataset(dset, sphere)
    assert all(pb.omega == 1.5 for pb in problems)
    assert all(pb.radiating_dof == "Heave" for pb in problems)


def test_problems_from_dataset_without_list_with_too_many_frequencies(sphere):
    dset = xr.Dataset(coords={'omega': 1.5, 'period': 1.5, 'radiating_dof': "Heave"})
    with pytest.raises(ValueError, match="at most one"):
        problems_from_dataset(dset, sphere)


#######################################################################
#                         Assemble dataframes                         #
#######################################################################

def test_assemble_dataframe(sphere, solver):
    pb_1 = cpt.DiffractionProblem(body=sphere, wave_direction=1.0, omega=1.0)
    res_1 = solver.solve(pb_1)
    df1 = cpt.assemble_dataframe([res_1])
    assert "diffraction_force" in df1
    assert "added_mass" not in df1

    pb_2 = cpt.RadiationProblem(body=sphere, radiating_dof="Heave", omega=1.0)
    res_2 = solver.solve(pb_2)
    df2 = cpt.assemble_dataframe([res_2])
    assert "added_mass" in df2
    assert "diffraction_force" not in df2

    df12 = cpt.assemble_dataframe([res_1, res_2])
    assert "diffraction_force" in df12
    assert "added_mass" in df12


def test_assemble_dataframe_with_infinite_free_surface(sphere, solver):
    pb = cpt.RadiationProblem(body=sphere, free_surface=np.inf)
    res = solver.solve(pb)
    df = cpt.assemble_dataframe([res])
    assert "free_surface" in df


#######################################################################
#                          Assemble matrices                          #
#######################################################################

def test_assemble_matrices(sphere, solver):
    pbs = [cpt.DiffractionProblem(body=sphere, wave_direction=1.0, omega=1.0),
           cpt.RadiationProblem(body=sphere, wave_direction=1.0, radiating_dof="Heave")]
    res = solver.solve_all(pbs)
    A, B, F = cpt.assemble_matrices(res)
    assert A.shape == (1, 1)
    assert A.dtype == np.float64
    assert B.shape == (1, 1)
    assert B.dtype == np.float64
    assert F.shape == (1,)
    assert F.dtype == np.complex128


def test_assemble_matrices_rad_only(sphere, solver):
    pbs = [cpt.RadiationProblem(body=sphere, wave_direction=1.0, radiating_dof="Heave")]
    res = solver.solve_all(pbs)
    A, B, F = cpt.assemble_matrices(res)
    assert A.shape == (1, 1)
    assert A.dtype == np.float64
    assert B.shape == (1, 1)
    assert B.dtype == np.float64
    assert F is None


def test_assemble_matrices_dif_only(sphere, solver):
    pbs = [cpt.DiffractionProblem(body=sphere, wave_direction=1.0, omega=1.0)]
    res = solver.solve_all(pbs)
    A, B, F = cpt.assemble_matrices(res)
    assert A is None
    assert B is None
    assert F.shape == (1,)
    assert F.dtype == np.complex128


def test_assemble_matrices_no_data():
    with pytest.raises(ValueError):
        cpt.assemble_matrices([])

#######################################################################
#                          Assemble dataset                           #
#######################################################################

def test_assemble_dataset(sphere, solver):
    pb_1 = cpt.DiffractionProblem(body=sphere, wave_direction=1.0, omega=1.0)
    res_1 = solver.solve(pb_1)
    ds1 = cpt.assemble_dataset([res_1])
    assert "diffraction_force" in ds1
    assert "added_mass" not in ds1

    pb_2 = cpt.RadiationProblem(body=sphere, radiating_dof="Heave", omega=1.0)
    res_2 = solver.solve(pb_2)
    ds2 = cpt.assemble_dataset([res_2])
    assert "added_mass" in ds2
    assert "diffraction_force" not in ds2

    ds12 = cpt.assemble_dataset([res_1, res_2])
    assert "diffraction_force" in ds12
    assert "added_mass" in ds12


def test_xarray_dataset_with_more_data():
    # Store some mesh data when several bodies in dataset
    bodies = [
        cpt.FloatingBody(cpt.mesh_sphere(radius=1, resolution=(3, 3)).immersed_part(), name="sphere_1"),
        cpt.FloatingBody(cpt.mesh_sphere(radius=2, resolution=(5, 3)).immersed_part(), name="sphere_2"),
        cpt.FloatingBody(cpt.mesh_sphere(radius=3, resolution=(7, 3)).immersed_part(), name="sphere_3"),
    ]
    for body in bodies:
        body.add_translation_dof(name="Heave")

    problems = [cpt.RadiationProblem(body=b, radiating_dof="Heave", omega=1.0) for b in bodies]
    results = cpt.BEMSolver().solve_all(problems)

    ds = cpt.assemble_dataset(results, mesh=True)
    assert 'nb_faces' in ds.coords
    assert set(ds.coords['nb_faces'].values) == set([b.mesh.nb_faces for b in bodies])


def test_variables_attrs(sphere, solver):
    pb = cpt.RadiationProblem(body=sphere, omega=1.0, radiating_dof="Heave")
    ds = cpt.assemble_dataset([solver.solve(pb)])
    assert 'long_name' in ds.omega.attrs
    assert 'long_name' in ds.wavenumber.attrs
    assert 'long_name' in ds.freq.attrs
    assert 'long_name' in ds.added_mass.attrs


def test_assemble_dataset_with_infinite_free_surface(caplog, sphere, solver):
    pb = cpt.RadiationProblem(body=sphere, free_surface=np.inf)
    res = solver.solve(pb)
    with caplog.at_level("WARNING"):
        ds = cpt.assemble_dataset([res])
    assert "ignored" in caplog.text
    assert len(ds) == 0


def test_assemble_dataset_with_nans(sphere):
    pb = cpt.DiffractionProblem(body=sphere, wave_direction=1.0, omega=1.0)
    res = pb.make_results_container(forces={dof: np.nan for dof in pb.influenced_dofs})
    ds = cpt.assemble_dataset([res])
    assert 'diffraction_force' in ds
    assert np.all(np.isnan(ds.diffraction_force.values))


def test_fill_dataset(sphere, solver):
    sphere.add_all_rigid_body_dofs()
    test_matrix = xr.Dataset(coords={'omega': [1.0, 2.0, 3.0], 'wave_direction': [0, np.pi/2], 'radiating_dof': ['Heave']})
    dataset = solver.fill_dataset(test_matrix, [sphere])
    assert dataset['added_mass'].data.shape == (3, 1, 6)
    assert dataset['Froude_Krylov_force'].data.shape == (3, 2, 6)


def test_fill_dataset_with_freqs(sphere, solver):
    f_range = np.linspace(0.1, 1, 3)
    test_matrix = xr.Dataset(coords={'freq': f_range, 'wave_direction': [0, np.pi/2], 'radiating_dof': ['Heave']})
    dataset = solver.fill_dataset(test_matrix, [sphere])
    np.testing.assert_allclose(dataset.coords['freq'], f_range)
    assert set(dataset.added_mass.dims) == {'freq', 'radiating_dof', 'influenced_dof'}
    assert set(dataset.freq.dims) == {'freq'}
    assert set(dataset.wavelength.dims) == {'freq'}
    assert set(dataset.omega.dims)      == {'freq'}
    assert set(dataset.period.dims)     == {'freq'}


def test_fill_dataset_with_wavenumbers(sphere, solver):
    k_range = np.linspace(1.0, 3.0, 3)
    test_matrix = xr.Dataset(coords={'wavenumber': k_range, 'wave_direction': [0, np.pi/2], 'radiating_dof': ['Heave']})
    dataset = solver.fill_dataset(test_matrix, [sphere])
    np.testing.assert_allclose(dataset.coords['wavenumber'], k_range)
    assert set(dataset.added_mass.dims) == {'wavenumber', 'radiating_dof', 'influenced_dof'}
    assert set(dataset.wavenumber.dims) == {'wavenumber'}
    assert set(dataset.wavelength.dims) == {'wavenumber'}
    assert set(dataset.omega.dims)      == {'wavenumber'}
    assert set(dataset.period.dims)     == {'wavenumber'}


def test_fill_dataset_with_periods(sphere, solver):
    T_range = np.linspace(1.0, 3.0, 3)
    test_matrix = xr.Dataset(coords={'period': T_range, 'wave_direction': [0, np.pi/2], 'radiating_dof': ['Heave']})
    dataset = solver.fill_dataset(test_matrix, [sphere])

    np.testing.assert_allclose(sorted(dataset.coords['period']), sorted(T_range))
    assert set(dataset.added_mass.dims) == {'period', 'radiating_dof', 'influenced_dof'}
    assert set(dataset.wavenumber.dims) == {'period'}
    assert set(dataset.wavelength.dims) == {'period'}
    assert set(dataset.omega.dims)      == {'period'}
    assert set(dataset.period.dims)     == {'period'}


def test_fill_dataset_with_wavenumbers_and_several_water_depths(sphere, solver):
    k_range = np.linspace(1.0, 3.0, 3)
    test_matrix = xr.Dataset(coords={
        'wavenumber': k_range, 'radiating_dof': ['Heave'], 'water_depth': [4.0, 6.0],
    })
    dataset = solver.fill_dataset(test_matrix, [sphere])

    np.testing.assert_allclose(dataset.coords['wavenumber'], k_range)
    assert set(dataset.added_mass.dims) == {'wavenumber', 'radiating_dof', 'influenced_dof', 'water_depth'}
    assert set(dataset.wavenumber.dims) == {'wavenumber'}
    assert set(dataset.wavelength.dims) == {'wavenumber'}
    assert set(dataset.omega.dims)      == {'wavenumber', 'water_depth'}
    assert set(dataset.period.dims)     == {'wavenumber', 'water_depth'}


@pytest.fixture
def broken_bem_solver():
    ref_gf = cpt.Delhommeau()
    class BrokenGreenFunction:
        def evaluate(self, m1, m2, fs, wd, wavenumber, *args, **kwargs):
            if wavenumber < 2.0:
                raise NotImplementedError("I'm potato")
            else:
                return ref_gf.evaluate(m1, m2, fs, wd, wavenumber, *args, **kwargs)
    broken_bem_solver = cpt.BEMSolver(green_function=BrokenGreenFunction())
    return broken_bem_solver


def test_failed_resolution_in_dataset(broken_bem_solver, sphere):
    test_matrix = xr.Dataset(coords={"wavenumber": np.linspace(0.1, 5.0, 5), "wave_direction": [0.0], "radiating_dof": ["Heave"]})
    ds = broken_bem_solver.fill_dataset(test_matrix, sphere)
    assert len(ds.wavenumber) == 5
    assert np.any(np.isnan(ds.added_mass.values))


#######################################################################
#                       separate_complex_values                       #
#######################################################################

def test_remove_complex_values():
    original_dataset = xr.Dataset(
        data_vars={
            'variable1': (['x', 'y'], np.random.rand(4, 5)),
            'variable2': (['x', 'y', 'z'], np.random.rand(4, 5, 3) + 1j * np.random.rand(4, 5, 3)),
            'variable3': (['y', 'z'], np.random.rand(5, 3) + 1j * np.random.rand(5, 3)),
        },
        coords={
            'x': (['x'], np.linspace(0, 10, 4))
        })

    real_dataset = separate_complex_values(original_dataset)
    assert np.allclose(real_dataset['variable3'].sel(complex='re').data,
                       np.real(original_dataset['variable3'].data),
                       atol=0.1)
    assert set(original_dataset.dims) == set(real_dataset.dims) - {'complex'}

    complex_dataset = merge_complex_values(real_dataset)
    assert set(original_dataset.dims) == set(complex_dataset.dims)



"""Tests for the mesh import of mesh from Hydrostar file format"""
from unittest import mock
import logging

import pytest
import numpy as np

import capytaine as cpt
from capytaine.io.mesh_loaders import load_HST


def parse_hst(file_content):
    # Workaround to avoid actually writing/reading a file on disk
    with mock.patch('capytaine.io.mesh_loaders._check_file', lambda foo: None):
        with mock.patch('capytaine.io.mesh_loaders.open', mock.mock_open(read_data=file_content)):
            return load_HST("mocked/filename")


def test_hst_type_0_triangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[1/2, 1/2, 0.0]]))


def test_hst_type_0_quadrangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[2/3, 2/3, 0.0]]))


def test_hst_type_1_quadrangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 1
        1 1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[1/2, 1/2, 0.0]]))


def test_hst_type_1_triangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 1
        1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[2/3, 2/3, 0.0]]))


def test_hst_implicit_coordinate_numbering():
    mesh = parse_hst("""
        COORDINATES
        0.0 0.0 0.0
        1.0 0.0 0.0
        1.0 1.0 0.0
        0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[1/2, 1/2, 0.0]]))


def test_hst_coordinate_numbering_error():
    with pytest.raises(ValueError):
        parse_hst("""
            COORDINATES
            1 0.0 0.0 0.0
            2 1.0 0.0 0.0
            3 1.0 1.0 0.0
            5 0.0 1.0 0.0
            ENDCOORDINATES
            """)

def test_hst_symmetry_1():
    mesh = parse_hst("""
        SYMMETRY 1
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half, cpt.Mesh)
    assert mesh.merged().nb_faces == 2


def test_hst_symmetry_2():
    mesh = parse_hst("""
        SYMMETRY 2
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half.half, cpt.Mesh)
    assert mesh.merged().nb_faces == 4


def test_hst_ignored_lines(caplog):
    with caplog.at_level(logging.WARNING):
        parse_hst("""
            USER mancellin
            NBBODY 1
            COORDINATES
            1 0.0 0.0 0.0
            2 1.0 0.0 0.0
            3 1.0 1.0 0.0
            4 0.0 1.0 0.0
            ENDCOORDINATES
            PANEL TYPE 0
            1 2 3 4
            ENDPANEL
                """)
    assert "HST mesh reader ignored" in caplog.text
    assert "USER mancellin" in caplog.text
    assert "NBBODY 1" in caplog.text



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# Capytaine should have been installed, for the version here, as well as for
# the autodoc
from capytaine import __version__


# -- Project information -----------------------------------------------------

project = 'capytaine'
copyright = '2024, Matthieu Ancellin'
author = 'Matthieu Ancellin'

# The short X.Y version
version = __version__.removesuffix(".dev")
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.extlinks',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_toolbox.collapse',
    'sphinx_proof',
    'sphinx_copybutton',
    'sphinxcontrib.mermaid',
]

extlinks = {
    "issue": ("https://github.com/capytaine/capytaine/issues/%s", "GH %s"),
    "pull": ("https://github.com/capytaine/capytaine/pull/%s", "PR %s"),
}

rst_prolog = f"""
.. |examples_folder| replace:: Github folder with examples for version {__version__}
.. _examples_folder: https://github.com/capytaine/capytaine/tree/v{__version__}/docs/user_manual/examples/
"""

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_theme_options = {
      'github_user': 'capytaine',
      'github_repo': 'capytaine',
      'github_banner': 'false',  # "Fork me on Github" banner
      'github_button': 'false',  # Button with e.g. stars number
      'logo_name': "Capytaine",
      'description': "a Python-based linear potential flow BEM solver",
      }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.

html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'capytainedoc'


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
# intersphinx_mapping = {'https://docs.python.org/': None}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Options for maths

math_number_all = True

# Options for napoleon

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_ivar = True

# Numbering theorems
numfig = True



"""Abstract structure of a class used to compute the Green function"""
# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

from abc import ABC, abstractmethod

import numpy as np

from capytaine.meshes.mesh_like_protocol import MeshLike


class GreenFunctionEvaluationError(Exception):
    pass


class AbstractGreenFunction(ABC):
    """Abstract method to evaluate the Green function."""

    floating_point_precision: str

    def _get_colocation_points_and_normals(self, mesh1, mesh2, adjoint_double_layer):
        if isinstance(mesh1, MeshLike):
            collocation_points = mesh1.faces_centers
            nb_collocation_points = mesh1.nb_faces
            if not adjoint_double_layer: # Computing the D matrix
                early_dot_product_normals = mesh2.faces_normals
            else: # Computing the K matrix
                early_dot_product_normals = mesh1.faces_normals

        elif isinstance(mesh1, np.ndarray) and mesh1.ndim == 2 and mesh1.shape[1] == 3:
            # This is used when computing potential or velocity at given points in postprocessing
            collocation_points = mesh1
            nb_collocation_points = mesh1.shape[0]
            if not adjoint_double_layer: # Computing the D matrix
                early_dot_product_normals = mesh2.faces_normals
            else: # Computing the K matrix
                early_dot_product_normals = np.zeros((nb_collocation_points, 3))
                # Dummy argument since this method is meant to be used either
                # - to compute potential, then only S is needed and early_dot_product_normals is irrelevant,
                # - to compute velocity, then the adjoint full gradient is needed and early_dot_product is False and this value is unused.
                # TODO: add an only_S argument and return an error here if (early_dot_product and not only_S)

        else:
            raise ValueError(f"Unrecognized first input for {self.__class__.__name__}.evaluate:\n{mesh1}")

        return collocation_points, early_dot_product_normals

    def _init_matrices(self, shape, early_dot_product):
        if self.floating_point_precision == "float32":
            dtype = "complex64"
        elif self.floating_point_precision == "float64":
            dtype = "complex128"
        else:
            raise NotImplementedError(
                    f"Unsupported floating point precision: {self.floating_point_precision}"
                    )

        S = np.zeros(shape, order="F", dtype=dtype)
        K = np.zeros((shape[0], shape[1], 1 if early_dot_product else 3), order="F", dtype=dtype)
        return S, K

    @abstractmethod
    def evaluate(self, mesh1, mesh2, free_surface, water_depth, wavenumber, adjoint_double_layer=True, early_dot_product=True):
        pass



from importlib import import_module
from scipy.optimize import brentq
import numpy as np

from capytaine.green_functions.abstract_green_function import AbstractGreenFunction, GreenFunctionEvaluationError


class LiangWuNoblesseGF(AbstractGreenFunction):
    """Wrapper for the Fortran implementation of the infinite depth Green function of [Liang, Wu, Noblesse, 2018].

    Uses the same implementation as Delhommeau() for the Rankine and reflected Rankine terms.

    """
    floating_point_precision = "float64"

    fortran_core = import_module("capytaine.green_functions.libs.Delhommeau_float64")
    tabulation_grid_shape_index = fortran_core.constants.liang_wu_noblesse
    exportable_settings = {'green_function': "LiangWuNoblesseGF"}

    # Dummy arrays that won't actually be used by the fortran code.
    prony_decomposition = np.zeros((1, 1))
    dispersion_relation_roots = np.empty(1)
    finite_depth_method_index = -9999
    tabulation_nb_integration_points = 1
    tabulated_r_range = np.empty(1)
    tabulated_z_range = np.empty(1)
    tabulated_integrals = np.empty(1)
    dummy_param = -999

    def __str__(self):
        return "LiangWuNoblesseGF()"

    def __repr__(self):
        return "LiangWuNoblesseGF()"

    def _repr_pretty_(self, p, cycle):
        p.text(self.__repr__())

    def evaluate(self,
                 mesh1, mesh2,
                 free_surface=0.0, water_depth=np.inf, wavenumber=1.0,
                 adjoint_double_layer=True, early_dot_product=True
                 ):

        if free_surface == np.inf or water_depth < np.inf:
            raise NotImplementedError("LiangWuNoblesseGF() is only implemented for infinite depth with a free surface")

        if wavenumber == np.inf:
            gf_singularities_index = self.fortran_core.constants.high_freq
        else:
            gf_singularities_index = self.fortran_core.constants.low_freq

        collocation_points, early_dot_product_normals = \
                self._get_colocation_points_and_normals(mesh1, mesh2, adjoint_double_layer)

        S, K = self._init_matrices(
            (collocation_points.shape[0], mesh2.nb_faces), early_dot_product=early_dot_product
        )

        self.fortran_core.matrices.build_matrices(
            collocation_points,  early_dot_product_normals,
            mesh2.vertices,      mesh2.faces + 1,
            mesh2.faces_centers, mesh2.faces_normals,
            mesh2.faces_areas,   mesh2.faces_radiuses,
            *mesh2.quadrature_points,
            wavenumber, np.inf,
            self.tabulation_nb_integration_points, self.tabulation_grid_shape_index,
            self.tabulated_r_range, self.tabulated_z_range, self.tabulated_integrals,
            self.dummy_param, self.prony_decomposition, self.dispersion_relation_roots,
            gf_singularities_index, adjoint_double_layer,
            S, K
        )

        if mesh1 is mesh2:
            self.fortran_core.matrices.add_diagonal_term(
                    mesh2.faces_centers, early_dot_product_normals, free_surface, K,
                    )

        if np.any(np.isnan(S)) or np.any(np.isnan(K)):
            raise GreenFunctionEvaluationError(
                    "Green function returned a NaN in the interaction matrix.\n"
                    "It could be due to overlapping panels.")

        if early_dot_product:
            K = K.reshape((collocation_points.shape[0], mesh2.nb_faces))

        return S, K


class FinGreen3D(AbstractGreenFunction):
    """Wrapper for the Fortran implementation of the finite depth Green function of [Liu et al.].

    Uses the same implementation as Delhommeau() for the Rankine and reflected Rankine terms.

    """
    floating_point_precision = "float64"

    fortran_core = import_module("capytaine.green_functions.libs.Delhommeau_float64")
    finite_depth_method_index = fortran_core.constants.fingreen3d_method
    gf_singularities_index = fortran_core.constants.low_freq

    # Dummy arrays that won't actually be used by the fortran code.
    prony_decomposition = np.zeros((1, 1))
    tabulation_nb_integration_points = 1
    tabulated_r_range = np.empty(1)
    tabulated_z_range = np.empty(1)
    tabulated_integrals = np.empty(1)
    dummy_param = -999

    def __init__(self, *, nb_dispersion_roots=200):
        self.nb_dispersion_roots = nb_dispersion_roots
        self.exportable_settings = {
            'green_function': "FinGreen3D",
            'nb_dispersion_roots': nb_dispersion_roots
        }

    def __str__(self):
        return f"FinGreen3D(nb_dispersion_roots={self.nb_dispersion_roots})"

    def __repr__(self):
        return f"FinGreen3D(nb_dispersion_roots={self.nb_dispersion_roots})"

    def _repr_pretty_(self, p, cycle):
        p.text(self.__repr__())

    def compute_dispersion_relation_roots(self, nk, wavenumber, depth):
        omega2_h_over_g = wavenumber*np.tanh(wavenumber*depth)*depth
        def root(i_root):
            return brentq(lambda y: omega2_h_over_g + y*np.tan(y), (2*i_root+1)*np.pi/2 + 1e-10, (2*i_root+2)*np.pi/2 - 1e-10)/depth
        return np.array([wavenumber] + [root(i_root) for i_root in range(nk-1)])

    def evaluate(self, mesh1, mesh2, free_surface, water_depth, wavenumber, adjoint_double_layer=True, early_dot_product=True):

        if free_surface == np.inf or water_depth == np.inf:
            raise NotImplementedError("FinGreen3D is only implemented for finite depth with a free surface.")
        if wavenumber == 0.0 or wavenumber == np.inf:
            raise NotImplementedError("FinGreen3D is only implemented for non-zero and non-infinite frequencies")

        dispersion_relation_roots = self.compute_dispersion_relation_roots(
            self.nb_dispersion_roots,
            wavenumber,
            water_depth
        )

        collocation_points, early_dot_product_normals = \
                self._get_colocation_points_and_normals(mesh1, mesh2, adjoint_double_layer)

        S, K = self._init_matrices(
            (collocation_points.shape[0], mesh2.nb_faces), early_dot_product=early_dot_product
        )

        self.fortran_core.matrices.build_matrices(
            collocation_points,  early_dot_product_normals,
            mesh2.vertices,      mesh2.faces + 1,
            mesh2.faces_centers, mesh2.faces_normals,
            mesh2.faces_areas,   mesh2.faces_radiuses,
            *mesh2.quadrature_points,
            wavenumber, water_depth,
            self.tabulation_nb_integration_points, self.dummy_param,
            self.tabulated_r_range, self.tabulated_z_range, self.tabulated_integrals,
            self.finite_depth_method_index, self.prony_decomposition, dispersion_relation_roots,
            self.gf_singularities_index, adjoint_double_layer,
            S, K
        )

        if mesh1 is mesh2:
            self.fortran_core.matrices.add_diagonal_term(
                    mesh2.faces_centers, early_dot_product_normals, free_surface, K,
                    )

        if np.any(np.isnan(S)) or np.any(np.isnan(K)):
            raise GreenFunctionEvaluationError(
                    "Green function returned a NaN in the interaction matrix.\n"
                    "It could be due to overlapping panels.")

        if early_dot_product:
            K = K.reshape((collocation_points.shape[0], mesh2.nb_faces))

        return S, K


class HAMS_GF(AbstractGreenFunction):
    floating_point_precision = "float64"

    exportable_settings = {'green_function': "HAMS_GF"}

    def __init__(self):
        self.infinite_depth_gf = LiangWuNoblesseGF()
        self.finite_depth_gf = FinGreen3D(nb_dispersion_roots=200)

    def __str__(self):
        return "HAMS_GF()"

    def __repr__(self):
        return "HAMS_GF()"

    def _repr_pretty_(self, p, cycle):
        p.text(self.__repr__())

    def evaluate(self, mesh1, mesh2, free_surface, water_depth, wavenumber, adjoint_double_layer=True, early_dot_product=True):
        if water_depth == np.inf:
            return self.infinite_depth_gf.evaluate(mesh1, mesh2, free_surface, water_depth, wavenumber, adjoint_double_layer, early_dot_product)
        else:
            return self.finite_depth_gf.evaluate(mesh1, mesh2, free_surface, water_depth, wavenumber, adjoint_double_layer, early_dot_product)



from typing import Tuple, Protocol, runtime_checkable
from numpy.typing import ArrayLike


@runtime_checkable
class MeshLike(Protocol):
    """Minimal API that a class describing a mesh should implement to be
    usable with the rest of Capytaine.

    The goal is two-fold:
       1. Use at runtime to identify a mesh for functions that behaves
       differently depending on the type of the input (e.g. Delhommeau().evaluate).
       2. Use as documentation for third-party mesh implementation.

    In the future, it could also be used for static typing.
    """
    vertices: ArrayLike
    faces: ArrayLike
    nb_vertices: int
    nb_faces: int
    faces_centers: ArrayLike
    faces_normals: ArrayLike
    faces_areas: ArrayLike
    faces_radiuses: ArrayLike
    quadrature_points: Tuple[ArrayLike, ArrayLike]

    def __short_str__(self) -> str:
        ...

    def extract_faces(self, faces_id):
        ...

    def join_meshes(*meshes, return_mask):
        ...

    def with_normal_vector_going_down(self, **kwargs):
        ...



"""Tests for the mesh import of mesh from WAMIT file format"""
from unittest import mock

import numpy as np
import pytest

import capytaine as cpt
from capytaine.io.mesh_loaders import load_GDF
from capytaine.io.mesh_writers import write_GDF


def parse_gdf(file_content):
    # Workaround to avoid actually writing/reading a file on disk
    with mock.patch('capytaine.io.mesh_loaders._check_file', lambda foo: None):
        with mock.patch('capytaine.io.mesh_loaders.open', mock.mock_open(read_data=file_content)):
            return load_GDF("mocked/filename")


def test_gdf_no_symmetry():
    mesh = parse_gdf("""Mesh description
        1.0  9.81
        0 0
        2
        0.0 0.0 -1.0
        1.0 0.0 -1.0
        1.0 1.0 -1.0
        0.0 1.0 -1.0
        1.0 0.0 -1.0
        2.0 0.0 -1.0
        2.0 1.0 -1.0
        1.0 1.0 -1.0
    """)
    assert isinstance(mesh, cpt.Mesh)
    assert mesh.nb_faces == 2


def test_gdf_no_symmetry_alternative_format():
    mesh = parse_gdf("""Mesh description
        1.0  9.81
        0 0
        2
        0.0 0.0 -1.0 1.0 0.0 -1.0 1.0 1.0 -1.0 0.0 1.0 -1.0
        1.0 0.0 -1.0 2.0 0.0 -1.0 2.0 1.0 -1.0 1.0 1.0 -1.0
    """)
    assert mesh.nb_faces == 2


def test_gdf_x_symmetry():
    mesh = parse_gdf("""Mesh description
        1.0  9.81
        1 0
        2
        0.0 0.0 -1.0
        1.0 0.0 -1.0
        1.0 1.0 -1.0
        0.0 1.0 -1.0
        1.0 0.0 -1.0
        2.0 0.0 -1.0
        2.0 1.0 -1.0
        1.0 1.0 -1.0
    """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert mesh.plane == cpt.yOz_Plane
    assert mesh.nb_faces == 4


def test_gdf_y_symmetry():
    mesh = parse_gdf("""Mesh description
        1.0  9.81
        0 1
        2
        0.0 0.0 -1.0
        1.0 0.0 -1.0
        1.0 1.0 -1.0
        0.0 1.0 -1.0
        1.0 0.0 -1.0
        2.0 0.0 -1.0
        2.0 1.0 -1.0
        1.0 1.0 -1.0
    """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert mesh.plane == cpt.xOz_Plane
    assert mesh.nb_faces == 4


def test_gdf_xy_symmetry():
    mesh = parse_gdf("""Mesh description
        1.0  9.81
        1 1
        2
        0.0 0.0 -1.0
        1.0 0.0 -1.0
        1.0 1.0 -1.0
        0.0 1.0 -1.0
        1.0 0.0 -1.0
        2.0 0.0 -1.0
        2.0 1.0 -1.0
        1.0 1.0 -1.0
    """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half, cpt.ReflectionSymmetricMesh)
    assert mesh.nb_faces == 8


########################################################


def test_write_and_load_gdf(tmpdir):
    mesh_path = tmpdir.join("temp_mesh.gdf")

    original_mesh = cpt.mesh_horizontal_cylinder()
    write_GDF(str(mesh_path), original_mesh.vertices, original_mesh.faces, ulen=1, gravity=9.81, isx=0, isy=0)

    read_mesh = load_GDF(str(mesh_path))

    np.testing.assert_allclose(
        read_mesh.vertices[read_mesh.faces],
        original_mesh.vertices[original_mesh.faces],
        atol=1e-6
        )

    np.testing.assert_allclose(
        read_mesh.faces_areas,
        original_mesh.faces_areas
        )

    np.testing.assert_allclose(
        read_mesh.faces_normals,
        original_mesh.faces_normals,
        atol=1e-6
        )

    np.testing.assert_allclose(
        read_mesh.faces_centers,
        original_mesh.faces_centers
        )

    np.testing.assert_allclose(
        read_mesh.faces_radiuses,
        original_mesh.faces_radiuses
        )

    np.testing.assert_allclose(
        read_mesh.volume,
        original_mesh.volume
        )



import pytest

import numpy as np
import xarray as xr

import capytaine as cpt
from capytaine.green_functions.abstract_green_function import GreenFunctionEvaluationError


###########
#  Setup  #
###########

def make_simple_body():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(direction=(1, 0, 0), name="Surge")
    return body

def make_symmetric_body():
    mesh = cpt.mesh_parallelepiped(resolution=(2, 2, 2), reflection_symmetry=True).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(direction=(1, 0, 0), name="Surge")
    return body

def make_lid_body():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=mesh.generate_lid())
    body.add_translation_dof(direction=(1, 0, 0), name="Surge")
    return body

BODIES = [make_simple_body, make_symmetric_body, make_lid_body]

SOLVERS = [cpt.BEMSolver(method='direct'), cpt.BEMSolver(method='indirect')]


####################################################
#  Define problem with zero or infinite frequency  #
####################################################
# Check that when a problem is defined with a zero or infinite frequency, the
# other representations of the frequency are consistent.

@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_define_problem_for_zero_freq(water_depth):
    pb = cpt.RadiationProblem(
        body=make_simple_body(),
        omega=0.0,
        water_depth=water_depth,
        radiating_dof="Surge"
    )
    assert float(pb.wavelength) == np.inf


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_define_problem_for_inf_freq(water_depth):
    pb = cpt.RadiationProblem(
        body=make_simple_body(),
        omega=np.inf,
        water_depth=water_depth,
        radiating_dof="Surge"
    )
    assert float(pb.wavelength) == 0.0


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_define_problem_for_zero_wavelength(water_depth):
    pb = cpt.RadiationProblem(
        body=make_simple_body(),
        wavelength=0.0,
        water_depth=water_depth,
        radiating_dof="Surge"
    )
    assert float(pb.period) == 0.0
    assert float(pb.wavenumber) == np.inf
    assert float(pb.freq) == np.inf


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_define_problem_for_inf_wavelength(water_depth):
    pb = cpt.RadiationProblem(
        body=make_simple_body(),
        wavelength=np.inf,
        water_depth=water_depth,
        radiating_dof="Surge"
    )
    assert float(pb.period) == np.inf
    assert float(pb.wavenumber) == 0.0


##############################################################
#  Solve radiation problems with zero or infinite frequency  #
##############################################################

@pytest.mark.parametrize("omega", [0.0, np.inf])
@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_radiation_bc_is_defined_at_limit_freq(omega, water_depth):
    pb = cpt.RadiationProblem(
        body=make_simple_body(),
        omega=omega,
        water_depth=water_depth,
    )
    from capytaine.tools.symbolic_multiplication import SymbolicMultiplication
    assert isinstance(pb.boundary_condition, SymbolicMultiplication)
    assert isinstance(pb.boundary_condition.value, np.ndarray)
    assert not np.any(np.isnan(pb.boundary_condition.value))

@pytest.mark.parametrize("body", BODIES)
@pytest.mark.parametrize("omega", [0.0, np.inf])
@pytest.mark.parametrize("solver", SOLVERS)
def test_solve_radiation_inf_depth(body, omega, solver):
    pb = cpt.RadiationProblem(
        body=body(),
        omega=omega,
        water_depth=np.inf
    )
    res = solver.solve(pb)
    assert isinstance(res.added_masses['Surge'], float)
    assert res.radiation_damping['Surge'] == 0.0


@pytest.mark.parametrize("body", BODIES)
@pytest.mark.parametrize("solver", SOLVERS)
def test_solve_radiation_fin_depth_infinite_frequency(body, solver):
    pb = cpt.RadiationProblem(
        body=body(),
        omega=np.inf,
        water_depth=5.0
    )
    res = solver.solve(pb)
    assert isinstance(res.added_masses['Surge'], float)
    assert res.radiation_damping['Surge'] == 0.0


@pytest.mark.parametrize("body", BODIES)
@pytest.mark.parametrize("solver", SOLVERS)
def test_solve_radiation_fin_depth_zero_frequencies(body, solver):
    pb = cpt.RadiationProblem(
        body=body(),
        omega=0.0,
        water_depth=5.0
    )
    with pytest.raises((NotImplementedError, GreenFunctionEvaluationError)):
        solver.solve(pb)

def test_no_warning_mesh_resolution_at_zero_wavelength(caplog):
    solver = SOLVERS[0]
    pb = cpt.RadiationProblem(
        body=make_simple_body(),
        wavelength=0.0,
    )
    with caplog.at_level("WARNING"):
        solver.solve(pb)
    assert "resolution " not in caplog.text


##############################################
#  Fails with correct error for diffraction  #
##############################################

@pytest.mark.parametrize("omega", [0.0, np.inf])
@pytest.mark.parametrize("water_depth", [10.0, np.inf])
@pytest.mark.parametrize("solver", SOLVERS)
def test_fail_to_solve_diffraction_problem_with_limit_freq(omega, water_depth, solver):
    pb = cpt.DiffractionProblem(
        body=make_simple_body(),
        omega=omega,
        water_depth=water_depth
    )
    with pytest.raises(ValueError, match="zero or infinite frequency"):
        solver.solve(pb)

@pytest.mark.parametrize("omega", [0.0, np.inf])
@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_diffraction_bc_is_nan_at_limit_freq(omega, water_depth):
    pb = cpt.DiffractionProblem(
        body=make_simple_body(),
        omega=omega,
        water_depth=water_depth
    )
    assert np.all(np.isnan(pb.boundary_condition))

@pytest.mark.parametrize("omega", [0.0, np.inf])
@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_froude_krylov_is_nan_at_limit_freq(omega, water_depth):
    from capytaine.bem.airy_waves import froude_krylov_force
    pb = cpt.DiffractionProblem(
        body=make_simple_body(),
        omega=omega,
        water_depth=water_depth
    )
    assert np.all(np.isnan(froude_krylov_force(pb)['Surge']))

##########################################
#  Fill datasets with limit frequencies  #
##########################################

def test_dataset_with_limit_frequency_including_radiation_and_diffraction(caplog):
    sphere = make_simple_body()
    test_matrix = xr.Dataset(coords={
        'omega': [0.0, 1.0, np.inf],
        'wave_direction': [0.0],
        'radiating_dof': list(sphere.dofs),
    })
    solver = cpt.BEMSolver()
    with caplog.at_level("INFO"):
        ds = solver.fill_dataset(test_matrix, sphere)
    assert np.all(np.isnan(ds.diffraction_force.sel(omega=0.0)))
    assert not np.any(np.isnan(ds.diffraction_force.sel(omega=1.0)))
    assert np.all(np.isnan(ds.diffraction_force.sel(omega=np.inf)))
    assert not np.any(np.isnan(ds.added_mass))
    assert not np.any(np.isnan(ds.radiation_damping))
    assert "Diffraction problems at zero or infinite frequency are not defined" in caplog.text



"""Tests for the matrices returned by cpt.Delhommeau() class"""

import pytest

import numpy as np
import capytaine as cpt


@pytest.mark.parametrize("adjoint_double_layer", [True, False])
def test_gradiant_G_shape(adjoint_double_layer):
    mesh = cpt.mesh_sphere(radius=1, center=(0, 0, 0), resolution=(10, 10)).immersed_part()
    S, gradG_1 = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False,adjoint_double_layer=adjoint_double_layer)
    assert gradG_1.shape == (mesh.nb_faces, mesh.nb_faces, 3)


def test_gradient_G_a_posteriori_scalar_product():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    S, gradG = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    S, K = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=True)

    # Scalar product with the faces normal vectors
    K_ = np.zeros_like(S)
    for i in range(mesh.nb_faces):
        for j in range(mesh.nb_faces):
            K_[i, j] = gradG[i, j, :] @ mesh.faces_normals[i, :]

    # Shorter implementiation of the exact same scalar products
    K__ = np.einsum('...jk,...k->...j', gradG, mesh.faces_normals)

    np.testing.assert_allclose(K, K_)
    np.testing.assert_allclose(K, K__)


def test_gradient_G_a_posteriori_scalar_product_directBIE():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    S, gradG = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False, adjoint_double_layer=False)
    S, K = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=True, adjoint_double_layer=False)

    # Scalar product with the faces normal vectors
    K_ = np.zeros_like(S)
    for i in range(mesh.nb_faces):
        for j in range(mesh.nb_faces):
            K_[i, j] = gradG[i, j, :] @ mesh.faces_normals[j, :]

    # Shorter implementiation of the exact same scalar products
    K__ = np.einsum('...k,...k->...', gradG, mesh.faces_normals)

    np.testing.assert_allclose(K, K_)
    np.testing.assert_allclose(K, K__)
    np.testing.assert_allclose(K_, K__)


def test_gradient_G_with_collocation_points():
    mesh = cpt.mesh_sphere(radius=1, center=(0, 0, 0), resolution=(10, 10)).immersed_part()
    _, gradG_1 = cpt.Delhommeau().evaluate(mesh.faces_centers, mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    _, gradG_2 = cpt.Delhommeau().evaluate(mesh.copy(), mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    np.testing.assert_allclose(gradG_1, gradG_2, atol=1e-10, rtol=1e-10)


@pytest.mark.parametrize("adjoint_double_layer", [True, False])
def test_gradient_G_diagonal_term(adjoint_double_layer):
    mesh = cpt.mesh_sphere(radius=1, center=(0, 0, 0), resolution=(10, 10)).immersed_part()
    # Passing two different Python objects
    _, gradG_1 = cpt.Delhommeau().evaluate(mesh.copy(), mesh, 0.0, np.inf, 1.0,
                                           early_dot_product=False, adjoint_double_layer=adjoint_double_layer)
    # Passing the same Python object
    _, gradG_2 = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0,
                                           early_dot_product=False, adjoint_double_layer=adjoint_double_layer)

    diag_normal = np.zeros_like(gradG_2)
    for i in range(mesh.nb_faces):
        diag_normal[i, i, :] = mesh.faces_normals[i, :]

    np.testing.assert_allclose(gradG_1, gradG_2 - 0.5*diag_normal, atol=1e-10, rtol=1e-10)


def test_a_posteriori_scalar_product_direct_method():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    S, gradG = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    D_ = np.zeros_like(S)
    for i in range(mesh.nb_faces):
        for j in range(mesh.nb_faces):
            D_[i, j] = gradG[i, j, :] @ mesh.faces_normals[j, :]
    D__ = np.einsum('...k,...k->...', gradG, mesh.faces_normals)
    np.testing.assert_allclose(D_, D__)



"""Tests for the computation of the Green function using Fortran routines."""

import pytest

import numpy as np
import capytaine as cpt

# def test_exponential_integral(x, y):
#     # Compare with Scipy implementation
#     from scipy.special import exp1
#     z = x + 1j*y
#
#     gf = cpt.Delhommeau()
#     def E1(z):
#         return np.exp(-z)*gf.fortran_core.delhommeau_integrals.exp_e1(z)
#
#     assert np.isclose(E1(z), exp1(z), rtol=1e-3)

#     # Test property (A3.5) of the function according to [Del, p.367].
#     if y != 0.0:
#         assert np.isclose(E1(np.conjugate(z)), np.conjugate(E1(z)), rtol=1e-3)


def test_infinite_depth_gf_singularities_variants():
    """Test how the values of the variants of the infinite depth Green function are related."""
    from scipy.special import j0
    gf = cpt.Delhommeau()
    def wave_part(*args):
        return gf.fortran_core.green_wave.wave_part_infinite_depth(
                *args[:3], *gf.all_tabulation_parameters, args[3])
    xi = np.array([0.0, 0.0, -0.5])
    xj = np.array([0.0, 1.0, -0.5])
    k = 2.0
    g_lf, nablag_lf = wave_part(xi, xj, k, 0)
    g_hf, nablag_hf = wave_part(xi, xj, k, 1)
    r = k * np.linalg.norm(xi[:2] - xj[:2])
    z = k * (xi[2] + xj[2])
    assert g_hf.real + 2*k/np.hypot(r, z) == pytest.approx(g_lf.real, rel=1e-4)
    assert g_lf.imag == pytest.approx(g_hf.imag)
    assert g_lf.imag == pytest.approx(2*np.pi*k*np.exp(z)*j0(r), rel=1e-4)


@pytest.mark.parametrize("gf_singularities", ["low_freq", "high_freq"])
def test_infinite_depth_gf_finite_differences(gf_singularities):
    """Compare the gradient of the Green function as returned by the code with the finite difference gradient"""
    # TODO: same in finite depth
    gf = cpt.Delhommeau(gf_singularities=gf_singularities)
    def wave_part(*args):
        return gf.fortran_core.green_wave.wave_part_infinite_depth(
                *args[:3], *gf.all_tabulation_parameters, gf.gf_singularities_fortran_enum[gf.gf_singularities])[0]
    def nabla_wave_part(*args):
        return gf.fortran_core.green_wave.wave_part_infinite_depth(
                *args[:3], *gf.all_tabulation_parameters, gf.gf_singularities_fortran_enum[gf.gf_singularities])[1]
    k = 1.0
    xi = np.array([0.0, 0.0, -0.5])
    xj = np.array([1.0, 1.0, -0.5])
    x_eps = np.array([1e-3, 0.0, 0.0])
    y_eps = np.array([0.0, 1e-3, 0.0])
    z_eps = np.array([0.0, 0.0, 1e-3])
    g = wave_part(xi, xj, k)
    g_x = (wave_part(xi + x_eps, xj, k) - g)/np.linalg.norm(x_eps)
    g_y = (wave_part(xi + y_eps, xj, k) - g)/np.linalg.norm(y_eps)
    g_z = (wave_part(xi + z_eps, xj, k) - g)/np.linalg.norm(z_eps)
    nabla_g = np.array([g_x, g_y, g_z])
    nablag_ref = nabla_wave_part(xi, xj, k)
    assert nablag_ref == pytest.approx(nabla_g, abs=1e-2)


def test_value_of_inf_depth_green_function_wave_part():
    gf = cpt.Delhommeau()
    def wave_part(*args):
        return gf.fortran_core.green_wave.wave_part_infinite_depth(
                *args[:3], *gf.all_tabulation_parameters, gf.fortran_core.constants.low_freq)
    k = 1.0
    xi = np.array([0.0, 0.0, -0.5])
    xj = np.array([1.0, 1.0, -0.5])
    g, nabla_g = wave_part(xi, xj, k)
    assert g == pytest.approx(-2.04 + 1.29j, abs=0.1)
    assert nabla_g == pytest.approx([0.25+0.88j,  0.25+0.88j, -0.88+1.29j], abs=0.1)


gfs = [
        cpt.Delhommeau(tabulation_nr=328, tabulation_nz=46, tabulation_nb_integration_points=251, tabulation_grid_shape="legacy"),
        cpt.XieDelhommeau(tabulation_nr=328, tabulation_nz=46, tabulation_nb_integration_points=251, tabulation_grid_shape="legacy"),
        # TODO: add more relevant Green functions
        ]

@pytest.mark.parametrize("gf", gfs)
def test_symmetry_of_the_green_function_infinite_depth(gf):
    k = 1.0
    xi = np.array([0.0, 0.0, -1.0])
    xj = np.array([1.0, 1.0, -2.0])
    g1, dg1 = gf.fortran_core.green_wave.wave_part_infinite_depth(
        xi, xj, k, *gf.all_tabulation_parameters, gf.gf_singularities_fortran_enum[gf.gf_singularities]
    )
    g2, dg2 = gf.fortran_core.green_wave.wave_part_infinite_depth(
        xj, xi, k, *gf.all_tabulation_parameters, gf.gf_singularities_fortran_enum[gf.gf_singularities]
    )
    assert g1 == pytest.approx(g2)
    assert dg1[0:2] == pytest.approx(-dg2[0:2])
    assert dg1[2] == pytest.approx(dg2[2])


def test_floating_point_precision():
    assert cpt.Delhommeau(floating_point_precision="float64").tabulated_integrals.dtype == np.float64
    assert cpt.Delhommeau(floating_point_precision="float32").tabulated_integrals.dtype == np.float32


def test_no_tabulation():
    mesh = cpt.mesh_sphere().immersed_part()
    tabed_gf = cpt.Delhommeau()
    untabed_gf = cpt.Delhommeau(tabulation_nr=0, tabulation_nz=0)
    assert np.allclose(untabed_gf.evaluate(mesh, mesh)[0], tabed_gf.evaluate(mesh, mesh)[0], atol=1e-2)


# def test_panels_near_free_surface():
#     area = 4.0
#     def gf_at_depth(d):
#         mesh = cpt.mesh_rectangle(size=(np.sqrt(area), np.sqrt(area)), resolution=(1, 1), center=(0, 0, d))
#         x = np.array([[0, 0, 0]])
#         S, _ = cpt.Delhommeau().evaluate(x, mesh, free_surface=0, water_depth=np.infty, wavenumber=3.0)
#         return S
#     gf_at_depth(0.0)
#     gf_at_depth(1e-9)
#     gf_at_depth(1e-5)
#     gf_at_depth(1e-2)


def test_exact_integration_of_rankine_terms():
    gf = cpt.Delhommeau()
    center = np.array([0.0, 0.0, 0.0])
    area = 1.0
    mesh = cpt.mesh_rectangle(size=(np.sqrt(area), np.sqrt(area)), center=center, resolution=(1, 1))
    rankine_g_once = gf.evaluate_rankine_only(center.reshape(1, -1), mesh)[0]
    mesh = cpt.mesh_rectangle(size=(np.sqrt(area), np.sqrt(area)), center=center, resolution=(11, 11))
    # Use odd number of panels to avoid having the center on a corner of a panel, which is not defined for the strong singularity of the derivative.
    rankine_g_parts = gf.evaluate_rankine_only(center.reshape(1, -1), mesh)[0]
    assert rankine_g_once == pytest.approx(rankine_g_parts.sum(), abs=1e-2)


def test_rankine_term_alone_or_in_zero_freq():
    import capytaine as cpt
    gf = cpt.Delhommeau()
    collocation_point = np.array([1.0, 0.0, -1.0]).reshape(1, -1)
    mirror_collocation_point = np.array([1.0, 0.0, 1.0]).reshape(1, -1)
    mesh = cpt.mesh_rectangle(size=(1.0, 1.0), center=(0.0, 0.0, -1.0), resolution=(1, 1))
    rankine_only = gf.evaluate_rankine_only(collocation_point, mesh)[0] + gf.evaluate_rankine_only(mirror_collocation_point, mesh)[0]
    zero_freq = gf.evaluate(collocation_point, mesh, water_depth=np.inf, wavenumber=0.0)[0]
    assert rankine_only == pytest.approx(zero_freq)


def test_rankine_term_alone_or_in_inf_freq():
    import capytaine as cpt
    gf = cpt.Delhommeau()
    collocation_point = np.array([1.0, 0.0, -1.0]).reshape(1, -1)
    mirror_collocation_point = np.array([1.0, 0.0, 1.0]).reshape(1, -1)
    mesh = cpt.mesh_rectangle(size=(1.0, 1.0), center=(0.0, 0.0, -1.0), resolution=(1, 1))
    rankine_only = gf.evaluate_rankine_only(collocation_point, mesh)[0] - gf.evaluate_rankine_only(mirror_collocation_point, mesh)[0]
    inf_freq = gf.evaluate(collocation_point, mesh, water_depth=np.inf, wavenumber=np.inf)[0]
    assert rankine_only == pytest.approx(inf_freq)


def test_exact_integration_with_nb_integration_points():
    # Test that the tabulation_nb_integration_points is actually taken into account.
    mesh = cpt.mesh_rectangle(center=(0, 0, -1), resolution=(1, 1))
    point = np.array([[100, 0, -1]])
    gf = cpt.Delhommeau(tabulation_nr=0, tabulation_nb_integration_points=11)
    val1 = gf.evaluate(point, mesh)
    gf = cpt.Delhommeau(tabulation_nr=0, tabulation_nb_integration_points=101)
    val2 = gf.evaluate(point, mesh)
    assert np.abs(val1[0] - val2[0]) > 1e-1


def test_low_freq_with_rankine_part_singularities():
    # TODO: also for not adjoint double layer
    gf1 = cpt.Delhommeau(gf_singularities="low_freq")
    gf2 = cpt.Delhommeau(gf_singularities="low_freq_with_rankine_part")
    point = np.array([[0.0, 0.0, -1.0]])
    area = 1.0
    wavenumber = 1.0
    mesh = cpt.mesh_rectangle(size=(np.sqrt(area), np.sqrt(area)), center=(0, 0, -1), resolution=(1, 1))
    S1, K1 = gf1.evaluate(point, mesh, wavenumber=wavenumber, early_dot_product=False)
    S2, K2 = gf2.evaluate(point, mesh, wavenumber=wavenumber, early_dot_product=False)
    assert S1 == S2
    assert np.allclose(K1[:, :, :2], K2[:, :, :2], atol=1e-12)
    assert np.allclose(K1[:, :, 2].imag, K2[:, :, 2].imag, atol=1e-12)
    assert not np.allclose(K1[0, 0, 2].real, K2[0, 0, 2].real, atol=1e-12)
    assert np.allclose(K1[0, 0, 2].real, K2[0, 0, 2].real, atol=1e-1)


def test_liangwunoblesse_wave_term():
    core = cpt.LiangWuNoblesseGF().fortran_core
    gf, gf_r = core.liangwunoblessewaveterm.havelockgf(1.0, -1.0)
    re_gf, _, im_gf, re_gf_r, im_gf_r = core.delhommeau_integrals(1.0, -1.0, 1000)
    assert gf == pytest.approx(- re_gf - im_gf*1j, rel=5e-2)
    assert gf_r == pytest.approx(- re_gf_r - im_gf_r*1j, rel=5e-2)


@pytest.mark.parametrize("wavenumber", [0.0, 1.0, np.inf])
def test_full_liangwunoblesse(wavenumber):
    gf = cpt.LiangWuNoblesseGF()
    ref_gf = cpt.Delhommeau()
    mesh = cpt.mesh_sphere().immersed_part()
    S, K = gf.evaluate(mesh, mesh, 0.0, np.inf, wavenumber)
    ref_S, ref_K = ref_gf.evaluate(mesh, mesh, 0.0, np.inf, wavenumber)
    np.testing.assert_allclose(S, ref_S, rtol=1e-2)
    np.testing.assert_allclose(K, ref_K, rtol=1e-2)



"""Variants of Delhommeau's method for the computation of the Green function."""
# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import os
import logging
from functools import lru_cache
from importlib import import_module

import numpy as np

from capytaine.tools.prony_decomposition import find_best_exponential_decomposition, PronyDecompositionFailure
from capytaine.tools.cache_on_disk import cache_directory

from capytaine.green_functions.abstract_green_function import AbstractGreenFunction, GreenFunctionEvaluationError

LOG = logging.getLogger(__name__)

_default_parameters = dict(
    tabulation_nr=676,
    tabulation_rmax=100.0,
    tabulation_nz=372,
    tabulation_zmin=-251.0,
    tabulation_nb_integration_points=1001,
    tabulation_grid_shape="scaled_nemoh3",
    finite_depth_method="newer",
    finite_depth_prony_decomposition_method="python",
    floating_point_precision="float64",
    gf_singularities="low_freq",
)


class Delhommeau(AbstractGreenFunction):
    """The Green function as implemented in Aquadyn and Nemoh.

    Parameters
    ----------
    tabulation_nr: int, optional
        Number of tabulation points for horizontal coordinate.
        If 0 is given, no tabulation is used at all.
        Default: 676
    tabulation_rmax: float, optional
        Maximum value of r range for the tabulation. (Minimum is zero.)
        Only used with the :code:`"scaled_nemoh3"` method.
        Default: 100.0
    tabulation_nz: int, optional
        Number of tabulation points for vertical coordinate.
        If 0 is given, no tabulation is used at all.
        Default: 372
    tabulation_zmin: float, optional
        Minimum value of z range for the tabulation. (Maximum is zero.)
        Only used with the :code:`"scaled_nemoh3"` method.
        Default: -251.0
    tabulation_nb_integration_points: int, optional
        Number of points for the numerical integration w.r.t. :math:`theta` of
        Delhommeau's integrals
        Default: 1000
    tabulation_grid_shape: string, optional
        Either :code:`"legacy"` or :code:`"scaled_nemoh3"`, which are the two
        methods currently implemented.
        Default: :code:`"scaled_nemoh3"`
    tabulation_cache_dir: str or None, optional
        Directory in which to save the tabulation file(s).
        If None, the tabulation is not saved on disk.
        Default: calls capytaine.tools.cache_on_disk.cache_directory(), which
        returns the value of the environment variable CAPYTAINE_CACHE_DIR if
        set, or else the default cache directory on your system.
    finite_depth_method: string, optional
        The method used to compute the finite depth Green function.
    finite_depth_prony_decomposition_method: string, optional
        The implementation of the Prony decomposition used to compute the
        finite water_depth Green function. Accepted values: :code:`'fortran'`
        for Nemoh's implementation (by default), :code:`'python'` for an
        experimental Python implementation.
        See :func:`find_best_exponential_decomposition`.
    floating_point_precision: string, optional
        Either :code:`'float32'` for single precision computations or
        :code:`'float64'` for double precision computations.
        Default: :code:`'float64'`.
    gf_singularities: string, optional
        Chose of the variant among the ways singularities can be extracted from
        the Green function. Currently only affects the infinite depth Green
        function.
        Default: "low_freq".

    Attributes
    ----------
    fortran_core:
        Compiled Fortran module with functions used to compute the Green
        function.
    tabulation_grid_shape_index: int
    gf_singularities_index: int
    finite_depth_method_index: int
        Integers passed to Fortran code to describe which method is used.
    tabulated_r_range: numpy.array of shape (tabulation_nr,) and type floating_point_precision
    tabulated_z_range: numpy.array of shape (tabulation_nz,) and type floating_point_precision
        Coordinates of the tabulation points.
    tabulated_integrals: numpy.array of shape (tabulation_nr, tabulation_nz, nb_tabulated_values) and type floating_point_precision
        Tabulated Delhommeau integrals.
    """

    dispersion_relation_roots = np.empty(1)  # dummy array


    def __init__(self, *,
                 tabulation_nr=_default_parameters["tabulation_nr"],
                 tabulation_rmax=_default_parameters["tabulation_rmax"],
                 tabulation_nz=_default_parameters["tabulation_nz"],
                 tabulation_zmin=_default_parameters["tabulation_zmin"],
                 tabulation_nb_integration_points=_default_parameters["tabulation_nb_integration_points"],
                 tabulation_grid_shape=_default_parameters["tabulation_grid_shape"],
                 tabulation_cache_dir=cache_directory(),
                 finite_depth_method=_default_parameters["finite_depth_method"],
                 finite_depth_prony_decomposition_method=_default_parameters["finite_depth_prony_decomposition_method"],
                 floating_point_precision=_default_parameters["floating_point_precision"],
                 gf_singularities=_default_parameters["gf_singularities"],
                 ):

        self.fortran_core = import_module(f"capytaine.green_functions.libs.Delhommeau_{floating_point_precision}")

        self.tabulation_grid_shape = tabulation_grid_shape
        fortran_enum = {
                'legacy': self.fortran_core.constants.legacy_grid,
                'scaled_nemoh3': self.fortran_core.constants.scaled_nemoh3_grid,
                              }
        self.tabulation_grid_shape_index = fortran_enum[tabulation_grid_shape]

        self.gf_singularities = gf_singularities
        self.gf_singularities_fortran_enum = {
                'high_freq': self.fortran_core.constants.high_freq,
                'low_freq': self.fortran_core.constants.low_freq,
                'low_freq_with_rankine_part': self.fortran_core.constants.low_freq_with_rankine_part,
                }

        self.finite_depth_method = finite_depth_method
        fortran_enum = {
                'legacy': self.fortran_core.constants.legacy_finite_depth,
                'newer': self.fortran_core.constants.newer_finite_depth,
                              }
        self.finite_depth_method_index = fortran_enum[finite_depth_method]

        self.floating_point_precision = floating_point_precision
        self.tabulation_nb_integration_points = tabulation_nb_integration_points

        self.tabulation_cache_dir = tabulation_cache_dir
        if tabulation_cache_dir is None:
            self._create_tabulation(tabulation_nr, tabulation_rmax,
                                    tabulation_nz, tabulation_zmin,
                                    tabulation_nb_integration_points)
        else:
            self._create_or_load_tabulation(tabulation_nr, tabulation_rmax,
                                            tabulation_nz, tabulation_zmin,
                                            tabulation_nb_integration_points,
                                            tabulation_cache_dir)

        self.finite_depth_prony_decomposition_method = finite_depth_prony_decomposition_method

        self.exportable_settings = {
            'green_function': self.__class__.__name__,
            'tabulation_nr': tabulation_nr,
            'tabulation_rmax': tabulation_rmax,
            'tabulation_nz': tabulation_nz,
            'tabulation_zmin': tabulation_zmin,
            'tabulation_nb_integration_points': tabulation_nb_integration_points,
            'tabulation_grid_shape': tabulation_grid_shape,
            'finite_depth_method': finite_depth_method,
            'finite_depth_prony_decomposition_method': finite_depth_prony_decomposition_method,
            'floating_point_precision': floating_point_precision,
            'gf_singularities': gf_singularities,
        }

        self._hash = hash(self.exportable_settings.values())

    def __hash__(self):
        return self._hash

    def __str__(self):
        # Print only the non-default values.
        to_be_printed = []
        for name, value in self.exportable_settings.items():
            if name in _default_parameters and value != _default_parameters[name]:
                to_be_printed.append(f"{name}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(to_be_printed)})"

    def __repr__(self):
        # Same as __str__ except all values are printed even when they are the
        # default value.
        to_be_printed = []
        for name, value in self.exportable_settings.items():
            if name in _default_parameters:
                to_be_printed.append(f"{name}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(to_be_printed)})"

    def _repr_pretty_(self, p, cycle):
        p.text(self.__repr__())

    def _create_or_load_tabulation(self, tabulation_nr, tabulation_rmax,
                                   tabulation_nz, tabulation_zmin,
                                   tabulation_nb_integration_points,
                                   tabulation_cache_dir):
        """This method either:
            - loads an existing tabulation saved on disk
            - generates a new tabulation with the data provided as argument and save it on disk.
        """

        # Normalize inputs
        tabulation_rmax = float(tabulation_rmax)
        tabulation_zmin = float(tabulation_zmin)

        filename = "tabulation_{}_{}_{}_{}_{}_{}_{}.npz".format(
            self.floating_point_precision, self.tabulation_grid_shape,
            tabulation_nr, tabulation_rmax, tabulation_nz, tabulation_zmin,
            tabulation_nb_integration_points
        )
        filepath = os.path.join(tabulation_cache_dir, filename)

        if os.path.exists(filepath):
            try:
                LOG.info("Loading tabulation from %s", filepath)
                loaded_arrays = np.load(filepath)
                self.tabulated_r_range = loaded_arrays["r_range"]
                self.tabulated_z_range = loaded_arrays["z_range"]
                self.tabulated_integrals = loaded_arrays["values"]
                return filename
            except (EOFError, FileNotFoundError, KeyError, ValueError):
                LOG.warning("Error loading tabulation from %s", filepath)

        self._create_tabulation(tabulation_nr, tabulation_rmax,
                                tabulation_nz, tabulation_zmin,
                                tabulation_nb_integration_points)
        LOG.debug("Saving tabulation in %s", filepath)
        np.savez_compressed(
            filepath, r_range=self.tabulated_r_range, z_range=self.tabulated_z_range,
            values=self.tabulated_integrals
        )
        return filename

    def _create_tabulation(self, tabulation_nr, tabulation_rmax,
                                   tabulation_nz, tabulation_zmin,
                                   tabulation_nb_integration_points):
        LOG.warning("Precomputing tabulation, it may take a few seconds.")
        self.tabulated_r_range = self.fortran_core.delhommeau_integrals.default_r_spacing(
                tabulation_nr, tabulation_rmax, self.tabulation_grid_shape_index
                )
        self.tabulated_z_range = self.fortran_core.delhommeau_integrals.default_z_spacing(
                tabulation_nz, tabulation_zmin, self.tabulation_grid_shape_index
                )
        self.tabulated_integrals = self.fortran_core.delhommeau_integrals.construct_tabulation(
                self.tabulated_r_range, self.tabulated_z_range, tabulation_nb_integration_points,
                )

    @property
    def all_tabulation_parameters(self):
        """An alias meant to pass to the Fortran functions all the parameters controlling the tabulation in a single item."""
        return (self.tabulation_nb_integration_points, self.tabulation_grid_shape_index,
                self.tabulated_r_range, self.tabulated_z_range, self.tabulated_integrals)

    @lru_cache(maxsize=128)
    def find_best_exponential_decomposition(self, dimensionless_wavenumber, *, method=None):
        """Compute the decomposition of a part of the finite water_depth Green function as a sum of exponential functions.

        Two implementations are available: the legacy Fortran implementation from Nemoh and a newer one written in Python.
        For some still unexplained reasons, the two implementations do not always give the exact same result.
        Until the problem is better understood, the Fortran implementation is the default one, to ensure consistency with Nemoh.
        The Fortran version is also significantly faster...

        Results are cached.

        Parameters
        ----------
        dimensionless_wavenumber: float
            dimensionless wavenumber: :math:`kh`
        method: str, optional
            "python" or "fortran". If not provided, uses self.finite_depth_prony_decomposition_method.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            the amplitude and growth rates of the exponentials
        """
        kh = dimensionless_wavenumber

        if method is None:
            method = self.finite_depth_prony_decomposition_method

        LOG.debug("\tCompute Prony decomposition in finite water_depth Green function "
                  "for dimensionless_wavenumber=%.2e", dimensionless_wavenumber)

        if method.lower() == 'python':
            if kh <= 0.1:
                raise NotImplementedError(
                    f"{self} cannot evaluate finite depth Green function "
                    f"for kh<0.1 (kh={kh})"
                    )
            elif kh < 1e5:
                # The function that will be approximated.
                sing_coef = (1 + np.tanh(kh))**2/(1 - np.tanh(kh)**2 + np.tanh(kh)/kh)
                def ref_function(x):
                    """The function that should be approximated by a sum of exponentials."""
                    return ((x + kh*np.tanh(kh)) * np.exp(x))/(x*np.sinh(x) - kh*np.tanh(kh)*np.cosh(x)) - sing_coef/(x - kh) - 2
            else:
                # Asymptotic approximation of the function for large kh, including infinite frequency
                def ref_function(x):
                    return -2/(1 + np.exp(-2*x)) + 2

            try:
                a, lamda = find_best_exponential_decomposition(ref_function, x_min=-0.1, x_max=20.0, n_exp_range=range(4, 31, 2), tol=1e-4)
                return np.stack([lamda, a])
            except PronyDecompositionFailure as e:
                raise GreenFunctionEvaluationError(
                    f"{self} cannot evaluate finite depth Green function "
                    f"for kh={dimensionless_wavenumber}"
                ) from e

        elif method.lower() == 'fortran':
            if kh > 1e5:
                raise NotImplementedError("Fortran implementation of the Prony decomposition does not support infinite frequency")
            omega2_h_over_g = kh*np.tanh(kh)
            nexp, pr_d = self.fortran_core.old_prony_decomposition.lisc(omega2_h_over_g, kh)
            return pr_d[0:2, :nexp]

        else:
            raise ValueError(f"Unrecognized name for the Prony decomposition method: {repr(method)}. Expected 'python' or 'fortran'.")

    def evaluate_rankine_only(self,
                              mesh1, mesh2,
                              adjoint_double_layer=True, early_dot_product=True
                              ):
        r"""Construct the matrices between mesh1 (that can also be a list of points)
        and mesh2 for a Rankine kernel.

        Parameters
        ----------
        mesh1: Mesh or CollectionOfMeshes or list of points
            mesh of the receiving body (where the potential is measured)
            if only S is wanted or early_dot_product is False, then only a list
            of points as an array of shape (n, 3) can be passed.
        mesh2: Mesh or CollectionOfMeshes
            mesh of the source body (over which the source distribution is integrated)
        adjoint_double_layer: bool, optional
            compute double layer for direct method (F) or adjoint double layer
            for indirect method (T) matrices (default: True)
        early_dot_product: boolean, optional
            if False, return K as a (n, m, 3) array storing ∫∇G
            if True, return K as a (n, m) array storing ∫∇G·n

        Returns
        -------
        tuple of real-valued numpy arrays
            the matrices :math:`S` and :math:`K`
        """
        collocation_points, early_dot_product_normals = \
                self._get_colocation_points_and_normals(mesh1, mesh2, adjoint_double_layer)

        S, K = self._init_matrices(
            (collocation_points.shape[0], mesh2.nb_faces), early_dot_product
        )

        self.fortran_core.matrices.add_rankine_term_only(
                collocation_points,  early_dot_product_normals,
                mesh2.vertices,      mesh2.faces + 1,
                mesh2.faces_centers, mesh2.faces_normals,
                mesh2.faces_areas,   mesh2.faces_radiuses,
                *mesh2.quadrature_points,
                adjoint_double_layer,
                S, K)

        if mesh1 is mesh2:
            self.fortran_core.matrices.add_diagonal_term(
                    mesh2.faces_centers, early_dot_product_normals, np.inf, K,
                    )

        S, K = np.real(S), np.real(K)

        if np.any(np.isnan(S)) or np.any(np.isnan(K)):
            raise GreenFunctionEvaluationError(
                    "Green function returned a NaN in the interaction matrix.\n"
                    "It could be due to overlapping panels.")

        if early_dot_product:
            K = K.reshape((collocation_points.shape[0], mesh2.nb_faces))

        return S, K


    def evaluate(self,
                 mesh1, mesh2,
                 free_surface=0.0, water_depth=np.inf, wavenumber=1.0,
                 adjoint_double_layer=True, early_dot_product=True
                 ):
        r"""The main method of the class, called by the engine to assemble the influence matrices.

        Parameters
        ----------
        mesh1: MeshLike or list of points
            mesh of the receiving body (where the potential is measured)
            if only S is wanted or early_dot_product is False, then only a list of points as an array of shape (n, 3) can be passed.
        mesh2: MeshLike
            mesh of the source body (over which the source distribution is integrated)
        free_surface: float, optional
            position of the free surface (default: :math:`z = 0`)
        water_depth: float, optional
            constant depth of water (default: :math:`+\infty`)
        wavenumber: float, optional
            wavenumber (default: 1.0)
        adjoint_double_layer: bool, optional
            compute double layer for direct method (F) or adjoint double layer for indirect method (T) matrices (default: True)
        early_dot_product: boolean, optional
            if False, return K as a (n, m, 3) array storing ∫∇G
            if True, return K as a (n, m) array storing ∫∇G·n

        Returns
        -------
        tuple of numpy arrays
            the matrices :math:`S` and :math:`K`
            the dtype of the matrix can be real or complex and depends on self.floating_point_precision
        """

        if free_surface == np.inf:  # No free surface, only a single Rankine source term
            if water_depth != np.inf:
                raise ValueError("When setting free_surface=inf, "
                                 "the water depth should also be infinite "
                                 f"(got water_depth={water_depth})")

            return self.evaluate_rankine_only(
                mesh1, mesh2,
                adjoint_double_layer=adjoint_double_layer,
                early_dot_product=early_dot_product,
            )

        # Main case:
        collocation_points, early_dot_product_normals = \
                self._get_colocation_points_and_normals(mesh1, mesh2, adjoint_double_layer)

        S, K = self._init_matrices(
            (collocation_points.shape[0], mesh2.nb_faces), early_dot_product
        )

        wavenumber = float(wavenumber)

        # Overrides gf_singularities setting in some specific cases, else use the class one.
        if water_depth < np.inf and self.finite_depth_method == 'legacy' and not self.gf_singularities == 'low_freq':
            gf_singularities = "low_freq"  # Reproduce legacy method behavior
            LOG.debug(
                f"Overriding gf_singularities='{self.gf_singularities}' because of finite_depth_method=='legacy'"
            )
        elif wavenumber == 0.0 and not self.gf_singularities == 'low_freq':
            gf_singularities = "low_freq"
            LOG.debug(
                f"Overriding gf_singularities='{self.gf_singularities}' because of wavenumber==0.0"
            )
        elif wavenumber == np.inf and not self.gf_singularities == 'high_freq':
            gf_singularities = "high_freq"
            LOG.debug(
                f"Overriding gf_singularities='{self.gf_singularities}' because of wavenumber==np.inf"
            )
        elif np.any(abs(mesh2.faces_centers[:, 2]) < 1e-6) and not self.gf_singularities == 'low_freq':
            gf_singularities = "low_freq"
            LOG.warning(
                f"Overriding gf_singularities='{self.gf_singularities}' because of free surface panels, "
                "which are currently only supported by gf_singularities='low_freq'"
            )
        else:
            gf_singularities = self.gf_singularities
        gf_singularities_index = self.gf_singularities_fortran_enum[gf_singularities]

        if water_depth == np.inf:
            prony_decomposition = np.zeros((1, 1))  # Dummy array that won't actually be used by the fortran code.
        else:
            prony_decomposition = self.find_best_exponential_decomposition(wavenumber*water_depth)

        # Main call to Fortran code
        self.fortran_core.matrices.build_matrices(
            collocation_points,  early_dot_product_normals,
            mesh2.vertices,      mesh2.faces + 1,
            mesh2.faces_centers, mesh2.faces_normals,
            mesh2.faces_areas,   mesh2.faces_radiuses,
            *mesh2.quadrature_points,
            wavenumber, water_depth,
            *self.all_tabulation_parameters,
            self.finite_depth_method_index, prony_decomposition, self.dispersion_relation_roots,
            gf_singularities_index, adjoint_double_layer,
            S, K
        )

        if mesh1 is mesh2:
            self.fortran_core.matrices.add_diagonal_term(
                    mesh2.faces_centers, early_dot_product_normals, free_surface, K,
                    )

        if np.any(np.isnan(S)) or np.any(np.isnan(K)):
            raise GreenFunctionEvaluationError(
                    "Green function returned a NaN in the interaction matrix.\n"
                    "It could be due to overlapping panels.")

        if early_dot_product:
            K = K.reshape((collocation_points.shape[0], mesh2.nb_faces))

        return S, K

################################

class XieDelhommeau(Delhommeau):
    """Legacy way to call the gf_singularities="low_freq" variant."""

    def __init__(self, **kwargs):
        super().__init__(gf_singularities="low_freq", **kwargs)



"""Tests for the mesh submodule: definition and transformation of collections of meshes and symmetric meshes."""

import pytest
import numpy as np
import capytaine as cpt


def test_collection_of_meshes():
    # Create some dummy meshes
    vertices = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], dtype=float)
    dummy_meshes = [cpt.Mesh(vertices, [[0, 1, 2, 3]])]
    for i in range(3):
        dummy_meshes.append(dummy_meshes[0].copy())
        dummy_meshes[i+1].translate_z(i+1)

    # A first collection from a list
    coll = cpt.CollectionOfMeshes(dummy_meshes[:2])
    assert coll.nb_submeshes == 2
    assert coll.nb_vertices == 8
    assert coll.nb_faces == 2
    assert coll[1].nb_faces == 1
    assert np.all(coll.faces_areas == np.asarray([1.0, 1.0]))
    assert np.all(coll.faces == np.asarray([[0, 1, 2, 3], [4, 5, 6, 7]]))
    assert np.all(coll.indices_of_mesh(1) == slice(1, 2))

    # A copy of the collection
    copy_coll = coll.copy()
    copy_coll.translate_x(1.0)  # Move
    assert copy_coll.nb_faces == 2
    assert np.all(copy_coll.vertices[:, 0] >= 1.0)  # Has moved

    # Another collection from an iterable
    other_coll = cpt.CollectionOfMeshes(iter(dummy_meshes))
    assert other_coll.nb_faces == 4
    assert np.all(other_coll.vertices[:, 0] <= 1.0)  # Did not move

    # A collection of collections
    big_coll = cpt.CollectionOfMeshes((copy_coll, other_coll))
    assert big_coll.nb_faces == 6
    assert big_coll.nb_submeshes == 2

    # Move one object in one of the sub-meshes
    copy_coll[1].translate_x(1.0)
    assert big_coll.vertices[:, 0].max() == 3.0

    # Merging the big collection
    merged = big_coll.merged()
    assert isinstance(merged, cpt.Mesh)


def test_collection():
    sphere = cpt.mesh_sphere(name="foo", center=(0, 0, -2))
    other_sphere = cpt.mesh_sphere(name="bar", center=(0, 0, 2))

    coll = cpt.CollectionOfMeshes([sphere, other_sphere], name="baz")
    coll2 = cpt.CollectionOfMeshes([sphere, other_sphere])

    assert coll == coll2
    assert hash(coll) == hash(coll2)

    assert coll[0] == coll2[0]

    coll.tree_view()

    clipped_coll = coll.keep_immersed_part(inplace=False)
    assert len(clipped_coll) == 1
    # assert clipped_coll.merged() == sphere.merged()

    assert np.allclose(sphere.center_of_mass_of_nodes, sphere.merged().center_of_mass_of_nodes)
    assert np.allclose(sphere.diameter_of_nodes, sphere.merged().diameter_of_nodes)


def test_join_reflection_symmetric_meshes():
    cylinder_1 = cpt.mesh_horizontal_cylinder(center=(0, 0, -2), reflection_symmetry=True)
    cylinder_2 = cpt.mesh_horizontal_cylinder(center=(0, 0, -4), reflection_symmetry=True)
    assert cylinder_1.plane == cylinder_2.plane
    both = cylinder_1.join_meshes(cylinder_2)
    assert isinstance(both, cpt.ReflectionSymmetricMesh)

def test_join_reflection_symmetric_meshes_masks():
    cylinder_1 = cpt.mesh_horizontal_cylinder(center=(0, 0, -2), reflection_symmetry=True)
    cylinder_2 = cpt.mesh_horizontal_cylinder(center=(0, 0, -4), reflection_symmetry=True)
    assert cylinder_1.plane == cylinder_2.plane
    both, masks = cylinder_1.join_meshes(cylinder_2, return_masks=True)
    assert isinstance(both, cpt.ReflectionSymmetricMesh)
    assert masks[0].shape == masks[1].shape == (both.nb_faces,)
    assert np.all(masks[0] ^ masks[1])
    assert sum(masks[0]) == cylinder_1.nb_faces
    assert sum(masks[1]) == cylinder_2.nb_faces

def test_join_nested_reflection_symmetric_meshes():
    cube_1 = cpt.mesh_parallelepiped(center=(0, 0, -2), reflection_symmetry=True)
    cube_2 = cpt.mesh_parallelepiped(center=(0, 0, -4), reflection_symmetry=True)
    assert cube_1.plane == cube_2.plane
    both = cube_1 + cube_2
    assert isinstance(both, cpt.ReflectionSymmetricMesh)
    assert isinstance(both.half, cpt.ReflectionSymmetricMesh)

def test_join_nested_reflection_symmetric_meshes_masks():
    cube_1 = cpt.mesh_parallelepiped(center=(0, 0, -2), reflection_symmetry=True)
    cube_2 = cpt.mesh_parallelepiped(center=(0, 0, -4), reflection_symmetry=True)
    assert cube_1.plane == cube_2.plane
    both, masks = cube_1.join_meshes(cube_2, return_masks=True)
    assert isinstance(both, cpt.ReflectionSymmetricMesh)
    assert isinstance(both.half, cpt.ReflectionSymmetricMesh)
    assert masks[0].shape == masks[1].shape == (both.nb_faces,)
    assert np.all(masks[0] ^ masks[1])
    assert sum(masks[0]) == cube_1.nb_faces
    assert sum(masks[1]) == cube_2.nb_faces

def test_join_axisymmetric_disks():
    """Given two axisymmetric meshes with the same axis, build a new axisymmetric mesh combining the two."""
    disk1 = cpt.mesh_disk(radius=1.0, center=(-1, 0, 0), resolution=(6, 6), normal=(1, 0, 0), axial_symmetry=True)
    disk2 = cpt.mesh_disk(radius=2.0, center=(1, 0, 0), resolution=(8, 6), normal=(1, 0, 0), axial_symmetry=True)
    joined = disk1.join_meshes(disk2, name="two_disks")
    assert isinstance(joined, cpt.AxialSymmetricMesh)
    assert joined.nb_faces == disk1.nb_faces + disk2.nb_faces
    joined.tree_view()

    disk3 = cpt.mesh_disk(radius=1.0, center=(0, 0, 0), resolution=(6, 4), axial_symmetry=True)
    with pytest.raises(AssertionError):
        disk1.join_meshes(disk3)

def test_join_axisymmetric_disks_masks():
    """Given two axisymmetric meshes with the same axis, build a new axisymmetric mesh combining the two."""
    disk1 = cpt.mesh_disk(radius=1.0, center=(-1, 0, 0), resolution=(6, 6), normal=(1, 0, 0), axial_symmetry=True)
    disk2 = cpt.mesh_disk(radius=2.0, center=(1, 0, 0), resolution=(8, 6), normal=(1, 0, 0), axial_symmetry=True)
    joined, masks = disk1.join_meshes(disk2, name="two_disks", return_masks=True)
    assert isinstance(joined, cpt.AxialSymmetricMesh)
    assert masks[0].shape == masks[1].shape == (joined.nb_faces,)
    assert np.all(masks[0] ^ masks[1])
    assert sum(masks[0]) == disk1.nb_faces
    assert sum(masks[1]) == disk2.nb_faces

def test_join_translational_cylinders():
    """Given two meshes with the same translation symmetry, join them into a single mesh with the same symmetry."""
    params = dict(length=10.0, reflection_symmetry=False, translation_symmetry=True, resolution=(0, 10, 10))
    mesh1 = cpt.mesh_horizontal_cylinder(radius=1.0, center=(0, 5, -5), **params)
    mesh2 = cpt.mesh_horizontal_cylinder(radius=2.0, center=(0, -5, -5), **params)
    joined = mesh1.join_meshes(mesh2)
    assert isinstance(joined, cpt.TranslationalSymmetricMesh)

def test_join_translational_cylinders_masks():
    """Given two meshes with the same translation symmetry, join them into a single mesh with the same symmetry."""
    params = dict(length=10.0, reflection_symmetry=False, translation_symmetry=True, resolution=(0, 10, 10))
    mesh1 = cpt.mesh_horizontal_cylinder(radius=1.0, center=(0, 5, -5), **params)
    mesh2 = cpt.mesh_horizontal_cylinder(radius=2.0, center=(0, -5, -5), **params)
    joined, masks = mesh1.join_meshes(mesh2, return_masks=True)
    assert isinstance(joined, cpt.TranslationalSymmetricMesh)
    assert masks[0].shape == masks[1].shape == (joined.nb_faces,)
    assert np.all(masks[0] ^ masks[1])
    assert sum(masks[0]) == mesh1.nb_faces
    assert sum(masks[1]) == mesh2.nb_faces

def test_mesh_splitting():
    mesh = cpt.mesh_sphere()

    splitted_mesh = mesh.sliced_by_plane(cpt.xOz_Plane.translated_y(0.5))
    assert isinstance(splitted_mesh, cpt.CollectionOfMeshes)
    assert splitted_mesh.merged() == mesh

    twice_splitted_mesh = splitted_mesh.sliced_by_plane(cpt.yOz_Plane)
    assert isinstance(twice_splitted_mesh[0], cpt.CollectionOfMeshes)
    assert isinstance(twice_splitted_mesh[1], cpt.CollectionOfMeshes)
    assert twice_splitted_mesh.merged() == mesh


def test_extract_one_face():
    sphere = cpt.mesh_sphere(axial_symmetry=True)
    assert sphere.submesh_containing_face(0) == (0, 0)
    assert sphere.submesh_containing_face(1) == (0, 1)
    assert sphere.submesh_containing_face(sphere[0].nb_faces-1) == (0, sphere[0].nb_faces-1)
    assert sphere.submesh_containing_face(sphere[0].nb_faces) == (1, 0)
    assert sphere.submesh_containing_face(sphere[0].nb_faces+1) == (1, 1)

    for i in [5, sphere[0].nb_faces+5]:
        assert np.allclose(sphere.extract_one_face(i).faces_centers[0], sphere.faces_centers[i])


def test_immersed_part():
    mesh = cpt.mesh_horizontal_cylinder(reflection_symmetry=True)
    assert mesh.immersed_part().merged() == mesh.merged().immersed_part()


def test_path_to_leaf():
    sphere_1 = cpt.mesh_sphere()
    sphere_2 = cpt.mesh_sphere()
    sphere_3 = cpt.mesh_sphere()
    sphere_4 = cpt.mesh_sphere()
    coll = cpt.CollectionOfMeshes([
        cpt.CollectionOfMeshes([sphere_1, sphere_2]),
        sphere_3, sphere_4])
    assert sphere_1.path_to_leaf() == [[]]
    assert coll.path_to_leaf() == [[0, 0], [0, 1], [1], [2]]


def test_empty_symmetric_mesh():
    mesh = cpt.mesh_sphere(center=(0, 0, 2), radius=0.5, axial_symmetry=True)
    empty_mesh = mesh.immersed_part()
    assert empty_mesh.first_slice.nb_faces == 0



import os
import logging
from glob import glob

import pytest

import capytaine as cpt
from capytaine.tools.cache_on_disk import cache_directory


def test_tabulation_caching_write_files():
    dir = cache_directory()
    cpt.Delhommeau(tabulation_nr=3, tabulation_nz=3)
    assert len(glob(os.path.join(dir, "tabulation*.npz"))) > 0


def test_tabulation_caching_write_files(caplog):
    dir = cache_directory()
    filename = cpt.Delhommeau()._create_or_load_tabulation(
        tabulation_nr=3,
        tabulation_rmax=1.0,
        tabulation_nz=3,
        tabulation_zmin=-1.0,
        tabulation_nb_integration_points=10,
        tabulation_cache_dir=dir
    )
    os.remove(os.path.join(dir, filename))
    with open(os.path.join(dir, filename), 'wb') as f:
        f.write(b"foooooo")
        ## Corrupting file
    with caplog.at_level(logging.WARNING):
        filename = cpt.Delhommeau()._create_or_load_tabulation(
            tabulation_nr=3,
            tabulation_rmax=1.0,
            tabulation_nz=3,
            tabulation_zmin=-1.0,
            tabulation_nb_integration_points=10,
            tabulation_cache_dir=dir
        )
    assert 'Error loading tabulation' in caplog.text



import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import capytaine as cpt

cpt.set_logging('INFO')

# Generate body
mesh = cpt.mesh_horizontal_cylinder(
    length=3.0, radius=1.0, center=(0, 0, -1.01), resolution=(5, 30, 15)
)
body = cpt.FloatingBody(mesh)
body.add_translation_dof(name="Heave")

test_matrix = xr.Dataset(coords={
    'omega': np.linspace(0.5, 4, 20),
    'radiating_dof': list(body.dofs.keys()),
    'water_depth': [np.inf, 4.0]
})

green_functions = [
        cpt.Delhommeau(gf_singularities="low_freq"),
        # cpt.Delhommeau(gf_singularities="high_freq"),  # For this problem, more difficult to converge
        cpt.HAMS_GF(),
        ]

data = []
for gf in green_functions:
    data.append(cpt.BEMSolver(green_function=gf).fill_dataset(test_matrix, body))

fig, axs = plt.subplots(2, 1, sharex=True, layout="constrained")
for gf, ds in zip(green_functions, data):
    ds['added_mass'].sel(water_depth=np.inf).plot(ax=axs[0], x='omega', linestyle="--", label=f"Infinite depth {gf}")
    ds['added_mass'].sel(water_depth=4.0).plot(ax=axs[0], x='omega', label=f"Finite depth {gf}")
    ds['radiation_damping'].sel(water_depth=np.inf).plot(ax=axs[1], x='omega', linestyle="--", label=f"Infinite depth {gf}")
    ds['radiation_damping'].sel(water_depth=4.0).plot(ax=axs[1], x='omega', label=f"Finite depth {gf}")
axs[0].set_title("Added mass")
axs[0].legend()
axs[1].set_title("Radiation damping")

plt.show()



"""Tools to use xarray Datasets as inputs and outputs.

.. todo:: This module could be tidied up a bit and some methods merged or
          uniformized.
"""
# Copyright (C) 2017-2025 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging
from datetime import datetime
from itertools import product
from collections import Counter
from typing import Sequence, List, Union

import numpy as np
import pandas as pd
import xarray as xr

from capytaine import __version__
from capytaine.bodies.bodies import FloatingBody
from capytaine.bem.problems_and_results import (
    LinearPotentialFlowProblem, DiffractionProblem, RadiationProblem,
    LinearPotentialFlowResult, _default_parameters)
from capytaine.post_pro.kochin import compute_kochin
from capytaine.io.bemio import dataframe_from_bemio


LOG = logging.getLogger(__name__)


#########################
#  Reading test matrix  #
#########################

def _unsqueeze_dimensions(data_array, dimensions=None):
    """Add scalar coordinates as dimensions of size 1."""
    if dimensions is None:
        dimensions = list(data_array.coords.keys())
    for dim in dimensions:
        if len(data_array.coords[dim].values.shape) == 0:
            data_array = xr.concat([data_array], dim=dim)
    return data_array


def problems_from_dataset(dataset: xr.Dataset,
                          bodies: Union[FloatingBody, Sequence[FloatingBody]],
                          ) -> List[LinearPotentialFlowProblem]:
    """Generate a list of problems from a test matrix.

    Parameters
    ----------
    dataset : xarray Dataset
        Test matrix containing the problems parameters.
    bodies : FloatingBody or list of FloatingBody
        The bodies on which the computations of the test matrix will be applied.
        They should all have different names.

    Returns
    -------
    list of LinearPotentialFlowProblem

    Raises
    ------
    ValueError
        if required fields are missing in the dataset
    """
    if isinstance(bodies, FloatingBody):
        bodies = [bodies]

    # Should be done before looking for `frequency_keys`, otherwise
    # frequencies provided as a scalar dimension will be skipped.
    dataset = _unsqueeze_dimensions(dataset)

    # SANITY CHECKS
    assert len(list(set(body.name for body in bodies))) == len(bodies), \
        "All bodies should have different names."

    # Warn user in case of key with unrecognized name (e.g. misspells)
    keys_in_dataset = set(dataset.dims)
    accepted_keys = {'wave_direction', 'radiating_dof', 'influenced_dof',
                     'body_name', 'omega', 'freq', 'period', 'wavelength', 'wavenumber',
                     'forward_speed', 'water_depth', 'rho', 'g', 'theta'}
    unrecognized_keys = keys_in_dataset.difference(accepted_keys)
    if len(unrecognized_keys) > 0:
        LOG.warning(f"Unrecognized key(s) in dataset: {unrecognized_keys}")

    if ("radiating_dof" not in keys_in_dataset) and ("wave_direction" not in keys_in_dataset):
        raise ValueError("Neither 'radiating_dof' nor 'wave_direction' has been provided in the dataset. "
                "No linear potential flow problem can be inferred.")

    frequency_keys = keys_in_dataset & {'omega', 'freq', 'period', 'wavelength', 'wavenumber'}
    if len(frequency_keys) > 1:
            raise ValueError("Setting problems requires at most one of the following: omega (angular frequency) OR freq (in Hz) OR period OR wavenumber OR wavelength.\n"
                             "Received {}".format(frequency_keys))
    # END SANITY CHECKS

    if len(frequency_keys) == 0:
        freq_type = "omega"
        freq_range = [_default_parameters['omega']]
    else:  # len(frequency_keys) == 1
        freq_type = list(frequency_keys)[0]  # Get the only item
        freq_range = dataset[freq_type].data

    water_depth_range = dataset['water_depth'].data if 'water_depth' in dataset else [_default_parameters['water_depth']]
    rho_range = dataset['rho'].data if 'rho' in dataset else [_default_parameters['rho']]
    g_range = dataset['g'].data if 'g' in dataset else [_default_parameters['g']]
    forward_speed_range = dataset['forward_speed'] if 'forward_speed' in dataset else [_default_parameters['forward_speed']]

    wave_direction_range = dataset['wave_direction'].data if 'wave_direction' in dataset else None
    radiating_dofs = dataset['radiating_dof'].data.astype(object) if 'radiating_dof' in dataset else None
    # astype(object) is meant to convert Numpy internal string type numpy.str_ to Python general string type.

    if 'body_name' in dataset:
        assert set(dataset['body_name'].data) <= {body.name for body in bodies}, \
            "Some body named in the dataset was not given as argument to `problems_from_dataset`."
        body_range = {body.name: body for body in bodies if body.name in dataset['body_name'].data}
        # Only the bodies listed in the dataset have been kept
    else:
        body_range = {body.name: body for body in bodies}

    problems = []
    if wave_direction_range is not None:
        for freq, wave_direction, water_depth, body_name, forward_speed, rho, g \
                in product(freq_range, wave_direction_range, water_depth_range, body_range,
                           forward_speed_range, rho_range, g_range):
            problems.append(
                DiffractionProblem(body=body_range[body_name], **{freq_type: freq},
                                   wave_direction=wave_direction, water_depth=water_depth,
                                   forward_speed=forward_speed, rho=rho, g=g)
            )

    if radiating_dofs is not None:
        for freq, radiating_dof, water_depth, body_name, forward_speed, rho, g \
                in product(freq_range, radiating_dofs, water_depth_range, body_range, forward_speed_range, rho_range, g_range):
            if forward_speed == 0.0:
                problems.append(
                    RadiationProblem(body=body_range[body_name], **{freq_type: freq},
                                     radiating_dof=radiating_dof, water_depth=water_depth,
                                     forward_speed=forward_speed, rho=rho, g=g)
                )
            else:
                if wave_direction_range is None:
                    LOG.warning("Dataset contains non-zero forward speed (forward_speed=%.2f) but no wave_direction has been provided. Wave direction of 0 rad (x-axis) has been assumed.", forward_speed)
                    wave_direction_range = [0.0]
                for wave_direction in wave_direction_range:
                    problems.append(
                        RadiationProblem(body=body_range[body_name], **{freq_type: freq},
                                         radiating_dof=radiating_dof, water_depth=water_depth,
                                         forward_speed=forward_speed, wave_direction=wave_direction,
                                         rho=rho, g=g)
                    )

    return sorted(problems)


########################
#  Dataframe creation  #
########################

def _detect_bemio_results(results, calling_function="_detect_bemio_results"):
    error_msg = (
        f"The function {calling_function} expected either a non-empty list of LinearPotentialFlowResult or a bemio.io object.\n"
        f"Instead, it received:\n{repr(results)}"
        )

    if hasattr(results, '__iter__'):
        if len(results) == 0:
            raise ValueError("Iterable provided to `assemble_dataset` is empty.")
        try:
            if 'capytaine' in results[0].__module__:
                bemio_import = False
            else:
                raise TypeError(error_msg)
        except:
            raise TypeError(error_msg)

    else:
        try:
            if 'bemio.io' in results.__module__:
                bemio_import = True
            else:
                raise TypeError(error_msg)
        except:
            raise TypeError(error_msg)

    return bemio_import


def assemble_dataframe(results, wavenumber=True, wavelength=True):
    if _detect_bemio_results(results, calling_function="assemble_dataframe"):
        return dataframe_from_bemio(results, wavenumber, wavelength) # TODO add hydrostatics

    records_list = [record for result in results for record in result.records]
    df = pd.DataFrame(records_list)

    all_dofs_in_order = list({k: None for r in results for k in r.body.dofs.keys()})
    # Using a dict above to remove duplicates while conserving ordering
    inf_dof_cat = pd.CategoricalDtype(categories=all_dofs_in_order)
    df["influenced_dof"] = df["influenced_dof"].astype(inf_dof_cat)
    if 'added_mass' in df.columns:
        rad_dof_cat = pd.CategoricalDtype(categories=all_dofs_in_order)
        df["radiating_dof"] = df["radiating_dof"].astype(rad_dof_cat)

    return df


######################
#  Dataset creation  #
######################

def _squeeze_dimensions(data_array, dimensions=None):
    """Remove dimensions if they are of size 1. The coordinates become scalar coordinates."""
    if dimensions is None:
        dimensions = data_array.dims
    for dim in dimensions:
        if len(data_array[dim]) == 1:
            data_array = data_array.squeeze(dim, drop=False)
    return data_array


def _dataset_from_dataframe(df: pd.DataFrame,
                            variables: Union[str, Sequence[str]],
                            dimensions: Sequence[str],
                            optional_dims: Sequence[str],
                            ) -> Union[xr.DataArray, xr.Dataset]:
    """Transform a pandas.Dataframe into a xarray.Dataset.

    Parameters
    ----------
    df: pandas.DataFrame
        the input dataframe
    variables: string or sequence of strings
        the variables that will be stored in the output dataset.
        If a single name is provided, a DataArray of this variable will be provided instead.
    dimensions: sequence of strings
        Names of dimensions the variables depends on.
        They will always appear as dimension in the output dataset.
    optional_dims: sequence of strings
        Names of dimensions the variables depends on.
        They will appears as dimension in the output dataset only if they have
        more than one different values.
    """
    df = df.drop_duplicates(optional_dims + dimensions)
    df = df.set_index(optional_dims + dimensions)
    da = df.to_xarray()[variables]
    da = _squeeze_dimensions(da, dimensions=optional_dims)
    return da


def hydrostatics_dataset(bodies: Sequence[FloatingBody]) -> xr.Dataset:
    """Create a dataset by looking for 'inertia_matrix' and 'hydrostatic_stiffness'
    for each of the bodies in the list passed as argument.
    """
    dataset = xr.Dataset()
    for body_property in ['inertia_matrix', 'hydrostatic_stiffness']:
        bodies_properties = {body.name: body.__getattribute__(body_property) for body in bodies if hasattr(body, body_property)}
        if len(bodies_properties) > 0:
            bodies_properties = xr.concat(bodies_properties.values(), pd.Index(bodies_properties.keys(), name='body_name'))
            bodies_properties = _squeeze_dimensions(bodies_properties, dimensions=['body_name'])
            dataset = xr.merge([dataset, {body_property: bodies_properties}], compat="no_conflicts", join="outer")
    return dataset


def kochin_data_array(results: Sequence[LinearPotentialFlowResult],
                      theta_range: Sequence[float],
                      **kwargs,
                      ) -> xr.Dataset:
    """Compute the Kochin function for a list of results and fills a dataset.

    .. seealso::
        :meth:`~capytaine.post_pro.kochin.compute_kochin`
            The present function is just a wrapper around :code:`compute_kochin`.
    """
    # TODO: this not very good to mix computation and data manipulation here...
    records = pd.DataFrame([
        dict(**result.problem._asdict(), theta=theta, kochin=kochin, kind=result.__class__.__name__)
        for result in results
        for theta, kochin in zip(theta_range.data,
                                 compute_kochin(result, theta_range, **kwargs))
    ])

    kochin_data = xr.Dataset()

    if "RadiationResult" in set(records['kind']):
        radiation = _dataset_from_dataframe(
            records[records['kind'] == "RadiationResult"],
            variables=['kochin'],
            dimensions=['omega', 'radiating_dof', 'theta'],
            optional_dims=['g', 'rho', 'body_name', 'water_depth', 'forward_speed', 'wave_direction']
        )
        kochin_data['kochin'] = radiation['kochin']

    if "DiffractionResult" in set(records['kind']):
        diffraction = _dataset_from_dataframe(
            records[records['kind'] == "DiffractionResult"],
            ['kochin'],
            dimensions=['omega', 'wave_direction', 'theta'],
            optional_dims=['g', 'rho', 'body_name', 'water_depth', 'forward_speed']
        )
        kochin_data['kochin_diffraction'] = diffraction['kochin']

    return kochin_data

VARIABLES_ATTRIBUTES = {
        "omega": {
            'long_name': 'Angular frequency',
            'units': 'rad/s',
            },
        "freq": {
            'long_name': 'Frequency',
            'units': 'Hz',
            },
        "period": {
            'long_name': 'Period',
            'units': 's',
            },
        "wavenumber": {
            'long_name': "Angular wavenumber",
            'units': 'rad/m',
            },
        "wavelength": {
            'long_name': "Wave length",
            'units': 'm',
            },
        "encounter_omega": {
            'long_name': "Encounter angular frequency",
            'units': 'rad/s',
            },
        "encounter_wave_direction": {
            'long_name': "Encounter wave direction",
            'units': 'rad',
            },
        "wave_direction": {
            'long_name': "Wave direction",
            'units': "rad"
            },
        "radiating_dof": {
            'long_name': 'Radiating DOF',
            },
        "influenced_dof": {
            'long_name': 'Influenced DOF',
            },
        "added_mass": {
            'long_name': 'Added mass',
            },
        "radiation_damping": {
            'long_name': 'Radiation damping',
            },
        "diffraction_force": {
            'long_name': "Diffraction force",
            },
        "Froude_Krylov_force": {
            'long_name': "Froude Krylov force",
            },
        }

def assemble_dataset(results,
                     omega=True, freq=True, wavenumber=True, wavelength=True, period=True,
                     mesh=False, hydrostatics=True, attrs=None) -> xr.Dataset:
    """Transform a list of :class:`LinearPotentialFlowResult` into a :class:`xarray.Dataset`.

    .. todo:: The :code:`mesh` option to store information on the mesh could be improved.
              It could store the full mesh in the dataset to ensure the reproducibility of
              the results.

    Parameters
    ----------
    results: list of LinearPotentialFlowResult or BEMIO dataset
        The results that will be read.
    omega: bool, optional
        If True, the coordinate 'omega' will be added to the output dataset.
    freq: bool, optional
        If True, the coordinate 'freq' will be added to the output dataset.
    wavenumber: bool, optional
        If True, the coordinate 'wavenumber' will be added to the output dataset.
    wavelength: bool, optional
        If True, the coordinate 'wavelength' will be added to the output dataset.
    period: bool, optional
        If True, the coordinate 'period' will be added to the output dataset.
    mesh: bool, optional
        If True, store some infos on the mesh in the output dataset.
    hydrostatics: bool, optional
        If True, store the hydrostatic data in the output dataset if they exist.
    attrs: dict, optional
        Attributes that should be added to the output dataset.
    """
    bemio_import = _detect_bemio_results(results, calling_function="assemble_dataset")

    records = assemble_dataframe(results)

    if bemio_import:
        main_freq_type = "omega"
    else:
        main_freq_type = Counter((res.provided_freq_type for res in results)).most_common(1)[0][0]

    if np.any(records["free_surface"] != 0.0):
        LOG.warning("Datasets only support cases with a free surface (free_surface=0.0).\n"
                    "Cases without a free surface (free_surface=inf) are ignored.\n"
                    "See also https://github.com/mancellin/capytaine/issues/88")
        records = records[records["free_surface"] == 0.0]

    if attrs is None:
        attrs = {}
    attrs['creation_of_dataset'] = datetime.now().isoformat()

    kinds_of_results = set(records['kind'])

    optional_dims = ['g', 'rho', 'body_name', 'water_depth', 'forward_speed']

    dataset = xr.Dataset()

    # RADIATION RESULTS
    if "RadiationResult" in kinds_of_results:
        radiation_cases = _dataset_from_dataframe(
            records[records['kind'] == "RadiationResult"],
            variables=['added_mass', 'radiation_damping'],
            dimensions=[main_freq_type, 'radiating_dof', 'influenced_dof'],
            optional_dims=optional_dims + ['wave_direction'])
        dataset = xr.merge([dataset, radiation_cases], compat="no_conflicts", join="outer")

    # DIFFRACTION RESULTS
    if "DiffractionResult" in kinds_of_results:
        diffraction_cases = _dataset_from_dataframe(
            records[records['kind'] == "DiffractionResult"],
            variables=['diffraction_force', 'Froude_Krylov_force'],
            dimensions=[main_freq_type, 'wave_direction', 'influenced_dof'],
            optional_dims=optional_dims)
        dataset = xr.merge([dataset, diffraction_cases], compat="no_conflicts", join="outer")
        dataset['excitation_force'] = dataset['Froude_Krylov_force'] + dataset['diffraction_force']

    # OTHER FREQUENCIES TYPES
    if omega and main_freq_type != "omega":
        omega_ds = _dataset_from_dataframe(
                records,
                variables=['omega'],
                dimensions=[main_freq_type],
                optional_dims=['g', 'water_depth'] if main_freq_type in {'wavelength', 'wavenumber'} else []
                )
        dataset.coords['omega'] = omega_ds['omega']

    if freq and main_freq_type != "freq":
        freq_ds = _dataset_from_dataframe(
                records,
                variables=['freq'],
                dimensions=[main_freq_type],
                optional_dims=['g', 'water_depth'] if main_freq_type in {'wavelength', 'wavenumber'} else []
                )
        dataset.coords['freq'] = freq_ds['freq']

    if period and main_freq_type != "period":
        period_ds = _dataset_from_dataframe(
                records,
                variables=['period'],
                dimensions=[main_freq_type],
                optional_dims=['g', 'water_depth'] if main_freq_type in {'wavelength', 'wavenumber'} else []
                )
        dataset.coords['period'] = period_ds['period']

    if wavenumber and main_freq_type != "wavenumber":
        wavenumber_ds = _dataset_from_dataframe(
                records,
                variables=['wavenumber'],
                dimensions=[main_freq_type],
                optional_dims=['g', 'water_depth'] if main_freq_type in {'period', 'omega'} else []
                )
        dataset.coords['wavenumber'] = wavenumber_ds['wavenumber']

    if wavelength and main_freq_type != "wavelength":
        wavelength_ds = _dataset_from_dataframe(
                records,
                variables=['wavelength'],
                dimensions=[main_freq_type],
                optional_dims=['g', 'water_depth'] if main_freq_type in {'period', 'omega'} else []
                )
        dataset.coords['wavelength'] = wavelength_ds['wavelength']

    if not all(records["forward_speed"] == 0.0):
        omegae_ds = _dataset_from_dataframe(
                records,
                variables=['encounter_omega'],
                dimensions=['forward_speed', 'wave_direction', main_freq_type],
                optional_dims=['g', 'water_depth'],
                )
        dataset.coords['encounter_omega'] = omegae_ds['encounter_omega']

        encounter_wave_direction_ds = _dataset_from_dataframe(
                records,
                variables=['encounter_wave_direction'],
                dimensions=['forward_speed', 'wave_direction', main_freq_type],
                optional_dims=[],
                )
        dataset.coords['encounter_wave_direction'] = encounter_wave_direction_ds['encounter_wave_direction']

    if mesh:
        if bemio_import:
            LOG.warning('Bemio data does not include mesh data. mesh=True is ignored.')
        else:
            # TODO: Store full mesh...
            bodies = list({result.body for result in results})  # Filter out duplicate bodies in the list of results
            nb_faces = {body.name: body.mesh.nb_faces for body in bodies}

            def name_or_str(c):
                return c.name if hasattr(c, 'name') else str(c)
            quad_methods = {body.name: name_or_str(body.mesh.quadrature_method) for body in bodies}

            if len(nb_faces) > 1:
                dataset.coords['nb_faces'] = ('body_name', [nb_faces[name] for name in dataset.coords['body_name'].data])
                dataset.coords['quadrature_method'] = ('body_name', [quad_methods[name] for name in dataset.coords['body_name'].data])
            else:
                def the_only(d):
                    """Return the only element of a 1-element dictionary"""
                    return next(iter(d.values()))
                dataset.coords['nb_faces'] = the_only(nb_faces)
                dataset.coords['quadrature_method'] = the_only(quad_methods)

    # HYDROSTATICS
    if hydrostatics:
        if bemio_import:
            LOG.warning('Bemio data import being used, hydrostatics=True is ignored.')
        else:
            bodies = list({result.body for result in results})
            dataset = xr.merge([dataset, hydrostatics_dataset(bodies)], compat="no_conflicts", join="outer")

    for var in set(dataset) | set(dataset.coords):
        if var in VARIABLES_ATTRIBUTES:
            dataset[var].attrs.update(VARIABLES_ATTRIBUTES[var])

    dataset.attrs.update(attrs)
    dataset.attrs['capytaine_version'] = __version__
    return dataset


def assemble_matrices(results):
    """Simplified version of assemble_dataset, returning only bare matrices.
    Meant mainly for teaching without introducing Xarray to beginers.

    Parameters
    ----------
    results: list of LinearPotentialFlowResult
        The results that will be read.

    Returns
    -------
    3-ple of (np.arrays or None)
        The added mass matrix, the radiation damping matrix and the excitation force.
        If the data are no available in the results, returns None instead.
    """

    ds = assemble_dataset(results)

    if "added_mass" in ds:
        A = np.atleast_2d(ds.added_mass.values.squeeze())
    else:
        A = None

    if "radiation_damping" in ds:
        B = np.atleast_2d(ds.radiation_damping.values.squeeze())
    else:
        B = None

    if "excitation_force" in ds:
        F = np.atleast_1d(ds.excitation_force.values.squeeze())
    else:
        F = None

    return A, B, F



################################
#  Handling of complex values  #
################################

def separate_complex_values(ds: xr.Dataset) -> xr.Dataset:
    """Return a new Dataset where complex-valued arrays of shape (...)
    have been replaced by real-valued arrays of shape (2, ...).

    .. seealso::
        :func:`merge_complex_values`
            The invert operation
    """
    ds = ds.copy()
    for variable in ds.data_vars:
        if ds[variable].dtype == complex:
            da = ds[variable]
            new_da = xr.DataArray(np.asarray((np.real(da).data, np.imag(da).data)),
                                  dims=('complex',) + da.dims)
            ds[variable] = new_da
            ds.coords['complex'] = ['re', 'im']
    return ds


def merge_complex_values(ds: xr.Dataset) -> xr.Dataset:
    """Return a new Dataset where real-valued arrays of shape (2, ...)
    have been replaced by complex-valued arrays of shape (...).

    .. seealso::
        :func:`separate_complex_values`
            The invert operation
    """
    if 'complex' in ds.coords:
        ds = ds.copy()
        for variable in ds.data_vars:
            if 'complex' in ds[variable].coords:
                da = ds[variable]
                new_dims = [d for d in da.dims if d != 'complex']
                new_da = xr.DataArray(da.sel(complex='re').data + 1j*da.sel(complex='im').data, dims=new_dims)
                ds[variable] = new_da
        ds = ds.drop_vars('complex')
    return ds


##################
#  Save dataset  #
##################

def save_dataset_as_netcdf(filename, dataset):
    """Save `dataset` as a NetCDF file with name (or path) `filename`"""
    ds = separate_complex_values(dataset)

    # Workaround https://github.com/capytaine/capytaine/issues/683
    ds['radiating_dof'] = ds['radiating_dof'].astype('str')
    ds['influenced_dof'] = ds['influenced_dof'].astype('str')

    # Make sure all strings are exported as strings and not Python objects
    encoding = {'radiating_dof': {'dtype': 'U'},
                'influenced_dof': {'dtype': 'U'}}

    ds.to_netcdf(filename, encoding=encoding)


def export_dataset(filename, dataset, format=None, **kwargs):
    """Save `dataset` into a format, provided by the `format` argument or inferred by the `filename`.

    Parameters
    ----------
    filename: str or Path
        Where to store the data
    dataset: xarray.Dataset
        Dataset, which is assumed to have been computed by Capytaine
    format: str, optional
        Format of output. Accepted values: "netcdf"
    **kwargs: optional
        Remaining argument are passed to the specific export function,
        such as ``save_dataset_as_netcdf``, ``export_to_wamit`` or ``write_dataset_as_tecplot_files``.

    Returns
    -------
    None
    """
    if (
            (format is not None and format.lower() == "netcdf") or
            (format is None and str(filename).endswith(".nc"))
            ):
        save_dataset_as_netcdf(filename, dataset, **kwargs)
    elif (
            (format is not None and format.lower() == "wamit")
            ):
        from capytaine.io.wamit import export_to_wamit
        export_to_wamit(dataset, filename, **kwargs)
    elif (
            (format is not None and format.lower() == "nemoh")
            ):
        from capytaine.io.legacy import write_dataset_as_tecplot_files
        write_dataset_as_tecplot_files(filename, dataset, **kwargs)
    else:
        raise ValueError("`export_dataset` could not infer export format based on filename or `format` argument.\n"
                         f"provided filename: {filename}\nprovided format: {format}")
